import datetime
import re

import pytest
from pytestskipmarkers.utils import platform

import salt.modules.config as configmod
import salt.modules.postgres as postgres
from salt.exceptions import SaltInvocationError
from tests.support.mock import MagicMock, Mock, call, patch

pytestmark = [
    pytest.mark.skip_unless_on_linux(reason="Only supported on Linux family"),
]


# 'md5' + md5('password' + 'username')
md5_pw = "md55a231fcdb710d73268c4f44283487ba2"

scram_pw = (
    "SCRAM-SHA-256$4096:wLr5nqC+3F+r7FdQPnB+nA==$"
    "0hn08ZdX8kirGaL4TM0j13digH9Wl365OOzCtAuF2pE=:"
    "LzAh/MGUdjYkdbDzcOKpfGwa3WwPUsyGcY+TEnSpcto="
)


@pytest.fixture
def get_test_privileges_list_function_csv():
    return """name
"{baruwatest=X/baruwatest,bayestest=r/baruwatest,baruwa=X*/baruwatest}"
"""


@pytest.fixture
def get_test_list_db_csv():
    return """Name,Owner,Encoding,Collate,Ctype,Access privileges,Tablespace
template1,postgres,LATIN1,en_US,en_US,"{=c/postgres,postgres=CTc/postgres}",pg_default
template0,postgres,LATIN1,en_US,en_US,"{=c/postgres,postgres=CTc/postgres}",pg_default
postgres,postgres,LATIN1,en_US,en_US,,pg_default
test_db,postgres,LATIN1,en_US,en_US,,pg_default
"""


@pytest.fixture
def get_test_list_schema_csv():
    return """name,owner,acl
public,postgres,"{postgres=UC/postgres,=UC/postgres}"
pg_toast,postgres,""
"""


@pytest.fixture
def get_test_list_language_csv():
    return "Name\ninternal\nc\nsql\nplpgsql\n"


@pytest.fixture
def get_test_privileges_list_table_csv():
    return """name
"{baruwatest=arwdDxt/baruwatest,bayestest=arwd/baruwatest,baruwa=a*r*w*d*D*x*t*/baruwatest}"
"""


@pytest.fixture
def get_test_privileges_list_group_csv():
    return "rolname,admin_option\nbaruwa,f\nbaruwatest2,t\nbaruwatest,f\n"


@pytest.fixture
def configure_loader_modules():
    return {
        postgres: {
            "__grains__": {"os_family": "Linux"},
            "__salt__": {
                "config.option": MagicMock(),
                "cmd.run_all": MagicMock(),
                "file.chown": MagicMock(),
                "file.remove": MagicMock(),
            },
        },
        configmod: {},
    }


def idfn(val):
    if val == md5_pw:
        return "md5_pw"
    if val == scram_pw:
        return "scram_pw"


@pytest.mark.parametrize(
    "role,password,verifier,method,result",
    [
        ("username", "password", md5_pw, "md5", True),
        ("another", "password", md5_pw, "md5", False),
        ("username", "another", md5_pw, "md5", False),
        ("username", md5_pw, md5_pw, "md5", True),
        ("username", "md5another", md5_pw, "md5", False),
        ("username", "password", md5_pw, True, True),
        ("another", "password", md5_pw, True, False),
        ("username", "another", md5_pw, True, False),
        ("username", md5_pw, md5_pw, True, True),
        ("username", "md5another", md5_pw, True, False),
        (None, "password", scram_pw, "scram-sha-256", True),
        (None, "another", scram_pw, "scram-sha-256", False),
        (None, scram_pw, scram_pw, "scram-sha-256", True),
        (None, "SCRAM-SHA-256$4096:AAAA$AAAA:AAAA", scram_pw, "scram-sha-256", False),
        (None, "SCRAM-SHA-256$foo", scram_pw, "scram-sha-256", False),
        (None, "password", "password", False, True),
        (None, "another", "password", False, False),
        (None, "password", "password", "foo", False),
        ("username", "password", md5_pw, "scram-sha-256", False),
        ("username", "password", scram_pw, "md5", False),
        # Code does not currently check role of pre-hashed md5 passwords
        pytest.param("another", md5_pw, md5_pw, "md5", False, marks=pytest.mark.xfail),
    ],
    ids=idfn,
)
def test_verify_password(role, password, verifier, method, result):
    if platform.is_fips_enabled() and (method == "md5" or verifier == md5_pw):
        pytest.skip("Test cannot run on a FIPS enabled platform")
    assert postgres._verify_password(role, password, verifier, method) == result


def test_has_privileges_with_function(get_test_privileges_list_function_csv):
    with patch(
        "salt.modules.postgres._run_psql",
        MagicMock(
            return_value={"retcode": 0, "stdout": get_test_privileges_list_function_csv}
        ),
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        ret = postgres.has_privileges(
            "baruwa",
            "awl",
            "function",
            "EXECUTE",
            grant_option=True,
            maintenance_db="db_name",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )

        assert ret is True

        query = (
            "COPY (SELECT rolname AS name "
            "FROM pg_catalog.pg_proc p "
            "JOIN pg_catalog.pg_namespace n "
            "ON n.oid = p.pronamespace "
            "JOIN pg_catalog.pg_roles r "
            "ON p.proowner = r.oid "
            "WHERE nspname = 'public' "
            "AND p.oid::regprocedure::text = 'awl' "
            "ORDER BY proname, proargtypes) TO STDOUT WITH CSV HEADER"
        )

        postgres._run_psql.assert_any_call(
            [
                "/usr/bin/pgsql",
                "--no-align",
                "--no-readline",
                "--no-psqlrc",
                "--no-password",
                "--username",
                "testuser",
                "--host",
                "testhost",
                "--port",
                "testport",
                "--dbname",
                "db_name",
                "-v",
                "datestyle=ISO,MDY",
                "-c",
                query,
            ],
            host="testhost",
            port="testport",
            password="testpassword",
            user="testuser",
            runas="user",
        )


def test__runpsql_with_timeout():
    cmd_run_mock = MagicMock()
    postgres_opts = {
        "config.option": configmod.option,
        "cmd.run_all": cmd_run_mock,
    }
    kwargs = {
        "reset_system_locale": False,
        "clean_env": True,
        "runas": "saltuser",
        "python_shell": False,
    }
    with patch.dict(postgres.__salt__, postgres_opts):
        with patch.dict(
            configmod.__opts__, {"postgres.timeout": 60, "postgres.pass": None}
        ):
            postgres._run_psql("fakecmd", runas="saltuser")
            cmd_run_mock.assert_called_with("fakecmd", timeout=60, **kwargs)
        with patch.dict(configmod.__opts__, {"postgres.pass": None}):
            postgres._run_psql("fakecmd", runas="saltuser")
            cmd_run_mock.assert_called_with("fakecmd", timeout=0, **kwargs)


def test__run_initdb_with_timeout():
    cmd_run_mock = MagicMock(return_value={})
    postgres_opts = {
        "config.option": configmod.option,
        "cmd.run_all": cmd_run_mock,
    }
    kwargs = {
        "clean_env": True,
        "runas": "saltuser",
        "python_shell": False,
    }
    cmd_str = "/fake/path --pgdata=fakename --username=saltuser --auth=password --encoding=UTF8"
    with patch.dict(postgres.__salt__, postgres_opts):
        with patch.object(postgres, "_find_pg_binary", return_value="/fake/path"):
            with patch.dict(
                configmod.__opts__, {"postgres.timeout": 60, "postgres.pass": None}
            ):
                postgres._run_initdb("fakename", runas="saltuser")
                cmd_run_mock.assert_called_with(cmd_str, timeout=60, **kwargs)
            with patch.dict(configmod.__opts__, {"postgres.pass": None}):
                postgres._run_initdb("fakename", runas="saltuser")
                cmd_run_mock.assert_called_with(cmd_str, timeout=0, **kwargs)


def test_run_psql():
    postgres._run_psql('echo "hi"')
    cmd = postgres.__salt__["cmd.run_all"]
    assert cmd.call_args[1]["runas"] == "postgres"


def test_db_alter():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        ret = postgres.db_alter(
            "dbname",
            user="testuser",
            host="testhost",
            port="testport",
            maintenance_db="maint_db",
            password="foo",
            tablespace="testspace",
            owner="otheruser",
            runas="foo",
        )
        assert ret is True

        postgres._run_psql.assert_has_calls(
            [
                call(
                    [
                        "/usr/bin/pgsql",
                        "--no-align",
                        "--no-readline",
                        "--no-psqlrc",
                        "--no-password",
                        "--username",
                        "testuser",
                        "--host",
                        "testhost",
                        "--port",
                        "testport",
                        "--dbname",
                        "maint_db",
                        "-c",
                        'ALTER DATABASE "dbname" OWNER TO "otheruser"',
                    ],
                    host="testhost",
                    user="testuser",
                    password="foo",
                    runas="foo",
                    port="testport",
                ),
                call(
                    [
                        "/usr/bin/pgsql",
                        "--no-align",
                        "--no-readline",
                        "--no-psqlrc",
                        "--no-password",
                        "--username",
                        "testuser",
                        "--host",
                        "testhost",
                        "--port",
                        "testport",
                        "--dbname",
                        "maint_db",
                        "-c",
                        'ALTER DATABASE "dbname" SET TABLESPACE "testspace"',
                    ],
                    host="testhost",
                    user="testuser",
                    password="foo",
                    runas="foo",
                    port="testport",
                ),
            ]
        )


def test_db_alter_owner_recurse():
    with patch(
        "salt.modules.postgres.owner_to", Mock(return_value={"retcode": None})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        postgres.db_alter(
            "dbname",
            user="testuser",
            host="testhost",
            port="testport",
            maintenance_db="maint_db",
            password="foo",
            tablespace="testspace",
            owner="otheruser",
            owner_recurse=True,
            runas="foo",
        )
        postgres.owner_to.assert_called_once_with(
            "dbname",
            "otheruser",
            user="testuser",
            host="testhost",
            port="testport",
            password="foo",
            runas="foo",
        )


def test_db_create():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        postgres.db_create(
            "dbname",
            user="testuser",
            host="testhost",
            port="testport",
            maintenance_db="maint_db",
            password="foo",
            tablespace="testspace",
            owner="otheruser",
            runas="foo",
        )

        postgres._run_psql.assert_called_once_with(
            [
                "/usr/bin/pgsql",
                "--no-align",
                "--no-readline",
                "--no-psqlrc",
                "--no-password",
                "--username",
                "testuser",
                "--host",
                "testhost",
                "--port",
                "testport",
                "--dbname",
                "maint_db",
                "-c",
                'CREATE DATABASE "dbname" WITH TABLESPACE = "testspace" '
                'OWNER = "otheruser"',
            ],
            host="testhost",
            user="testuser",
            password="foo",
            runas="foo",
            port="testport",
        )


def test_db_create_empty_string_param():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        postgres.db_create(
            "dbname",
            lc_collate="",
            encoding="utf8",
            user="testuser",
            host="testhost",
            port=1234,
            maintenance_db="maint_db",
            password="foo",
        )

        postgres._run_psql.assert_called_once_with(
            [
                "/usr/bin/pgsql",
                "--no-align",
                "--no-readline",
                "--no-psqlrc",
                "--no-password",
                "--username",
                "testuser",
                "--host",
                "testhost",
                "--port",
                "1234",
                "--dbname",
                "maint_db",
                "-c",
                "CREATE DATABASE \"dbname\" WITH ENCODING = 'utf8' LC_COLLATE = ''",
            ],
            host="testhost",
            password="foo",
            port=1234,
            runas=None,
            user="testuser",
        )


def test_db_create_with_trivial_sql_injection():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        pytest.raises(
            SaltInvocationError,
            postgres.db_create,
            "dbname",
            lc_collate="foo' ENCODING='utf8",
        )


def test_db_exists(get_test_list_db_csv):
    with patch(
        "salt.modules.postgres._run_psql",
        Mock(return_value={"retcode": 0, "stdout": get_test_list_db_csv}),
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        ret = postgres.db_exists(
            "test_db",
            user="testuser",
            host="testhost",
            port="testport",
            maintenance_db="maint_db",
            password="foo",
            runas="foo",
        )
        assert ret is True


def test_db_list(get_test_list_db_csv):
    with patch(
        "salt.modules.postgres._run_psql",
        Mock(return_value={"retcode": 0, "stdout": get_test_list_db_csv}),
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        ret = postgres.db_list(
            user="testuser",
            host="testhost",
            port="testport",
            maintenance_db="maint_db",
            password="foo",
            runas="foo",
        )
        assert ret == {
            "test_db": {
                "Encoding": "LATIN1",
                "Ctype": "en_US",
                "Tablespace": "pg_default",
                "Collate": "en_US",
                "Owner": "postgres",
                "Access privileges": "",
            },
            "template1": {
                "Encoding": "LATIN1",
                "Ctype": "en_US",
                "Tablespace": "pg_default",
                "Collate": "en_US",
                "Owner": "postgres",
                "Access privileges": "{=c/postgres,postgres=CTc/postgres}",
            },
            "template0": {
                "Encoding": "LATIN1",
                "Ctype": "en_US",
                "Tablespace": "pg_default",
                "Collate": "en_US",
                "Owner": "postgres",
                "Access privileges": "{=c/postgres,postgres=CTc/postgres}",
            },
            "postgres": {
                "Encoding": "LATIN1",
                "Ctype": "en_US",
                "Tablespace": "pg_default",
                "Collate": "en_US",
                "Owner": "postgres",
                "Access privileges": "",
            },
        }


def test_db_remove():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        postgres.db_remove(
            "test_db",
            user="testuser",
            host="testhost",
            port="testport",
            maintenance_db="maint_db",
            password="foo",
            runas="foo",
        )

        calls = (
            call(
                [
                    "/usr/bin/pgsql",
                    "--no-align",
                    "--no-readline",
                    "--no-psqlrc",
                    "--no-password",
                    "--username",
                    "testuser",
                    "--host",
                    "testhost",
                    "--port",
                    "testport",
                    "--dbname",
                    "maint_db",
                    "-c",
                    'REVOKE CONNECT ON DATABASE "test_db" FROM public;',
                ],
                host="testhost",
                password="foo",
                port="testport",
                runas="foo",
                user="testuser",
            ),
            call(
                [
                    "/usr/bin/pgsql",
                    "--no-align",
                    "--no-readline",
                    "--no-psqlrc",
                    "--no-password",
                    "--username",
                    "testuser",
                    "--host",
                    "testhost",
                    "--port",
                    "testport",
                    "--dbname",
                    "maint_db",
                    "-c",
                    "SELECT pid, pg_terminate_backend(pid) FROM pg_stat_activity"
                    " WHERE datname = 'test_db' AND pid <> pg_backend_pid();",
                ],
                host="testhost",
                password="foo",
                port="testport",
                runas="foo",
                user="testuser",
            ),
            call(
                [
                    "/usr/bin/pgsql",
                    "--no-align",
                    "--no-readline",
                    "--no-psqlrc",
                    "--no-password",
                    "--username",
                    "testuser",
                    "--host",
                    "testhost",
                    "--port",
                    "testport",
                    "--dbname",
                    "maint_db",
                    "-c",
                    'DROP DATABASE "test_db";',
                ],
                host="testhost",
                password="foo",
                port="testport",
                runas="foo",
                user="testuser",
            ),
        )

        postgres._run_psql.assert_has_calls(calls, any_order=True)


def test_group_create():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")), patch(
        "salt.modules.postgres.user_exists", Mock(return_value=False)
    ):
        postgres.group_create(
            "testgroup",
            user="testuser",
            host="testhost",
            port="testport",
            maintenance_db="maint_db",
            password="foo",
            createdb=False,
            encrypted=False,
            superuser=False,
            replication=False,
            rolepassword="testrolepass",
            groups="testgroup",
            runas="foo",
        )
        # postgres._run_psql.call_args[0][0] will contain the list of CLI args.
        # The first 14 elements of this list are initial args used in all (or
        # virtually all) commands run through _run_psql(), so the actual SQL
        # query will be in the 15th argument.
        assert postgres._run_psql.call_args[0][0][14].startswith("CREATE ROLE")


def test_group_remove():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")), patch(
        "salt.modules.postgres.user_exists", Mock(return_value=True)
    ):
        postgres.group_remove(
            "testgroup",
            user="testuser",
            host="testhost",
            port="testport",
            maintenance_db="maint_db",
            password="foo",
            runas="foo",
        )
        postgres._run_psql.assert_called_once_with(
            [
                "/usr/bin/pgsql",
                "--no-align",
                "--no-readline",
                "--no-psqlrc",
                "--no-password",
                "--username",
                "testuser",
                "--host",
                "testhost",
                "--port",
                "testport",
                "--dbname",
                "maint_db",
                "-c",
                'DROP ROLE "testgroup"',
            ],
            host="testhost",
            user="testuser",
            password="foo",
            runas="foo",
            port="testport",
        )


def test_group_update():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch(
        "salt.modules.postgres.role_get",
        Mock(return_value={"superuser": False}),
    ):
        postgres.group_update(
            "testgroup",
            user='"testuser"',
            host="testhost",
            port="testport",
            maintenance_db="maint_db",
            password="foo",
            createdb=False,
            encrypted=False,
            replication=False,
            rolepassword="test_role_pass",
            groups="testgroup",
            runas="foo",
        )
        # postgres._run_psql.call_args[0][0] will contain the list of CLI args.
        # The first 14 elements of this list are initial args used in all (or
        # virtually all) commands run through _run_psql(), so the actual SQL
        # query will be in the 15th argument.
        assert re.match(
            'ALTER.* "testgroup" .* UNENCRYPTED PASSWORD',
            postgres._run_psql.call_args[0][0][14],
        )


def test_user_create():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.modules.postgres.user_exists", Mock(return_value=False)):
        postgres.user_create(
            "testuser",
            user="testuser",
            host="testhost",
            port="testport",
            maintenance_db="maint_test",
            password="test_pass",
            login=True,
            createdb=False,
            createroles=False,
            encrypted=False,
            superuser=False,
            replication=False,
            rolepassword="test_role_pass",
            valid_until="2042-07-01",
            groups="test_groups",
            runas="foo",
        )
        # postgres._run_psql.call_args[0][0] will contain the list of CLI args.
        # The first 14 elements of this list are initial args used in all (or
        # virtually all) commands run through _run_psql(), so the actual SQL
        # query will be in the 15th argument.
        call = postgres._run_psql.call_args[0][0][14]
        assert re.match('CREATE ROLE "testuser"', call)
        for i in (
            "INHERIT",
            "NOCREATEDB",
            "NOCREATEROLE",
            "NOSUPERUSER",
            "NOREPLICATION",
            "LOGIN",
            "UNENCRYPTED",
            "PASSWORD",
            "VALID UNTIL",
        ):
            assert i in call, f"{i} not in {call}"


def test_user_exists():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.modules.postgres.version", Mock(return_value="9.1")), patch(
        "salt.modules.postgres.psql_query",
        Mock(
            return_value=[
                {
                    "name": "test_user",
                    "superuser": "t",
                    "inherits privileges": "t",
                    "can create roles": "t",
                    "can create databases": "t",
                    "can update system catalogs": "t",
                    "can login": "t",
                    "replication": None,
                    "password": "test_password",
                    "connections": "-1",
                    "groups": "",
                    "expiry time": "",
                    "defaults variables": None,
                }
            ]
        ),
    ):
        ret = postgres.user_exists(
            "test_user",
            user="test_user",
            host="test_host",
            port="test_port",
            maintenance_db="maint_db",
            password="test_password",
            runas="foo",
        )
        assert ret is True


def test_user_list():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.modules.postgres.version", Mock(return_value="9.1")), patch(
        "salt.modules.postgres.psql_query",
        Mock(
            return_value=[
                {
                    "name": "test_user",
                    "superuser": "t",
                    "inherits privileges": "t",
                    "can create roles": "t",
                    "can create databases": "t",
                    "can update system catalogs": "t",
                    "can login": "t",
                    "replication": None,
                    "connections": "-1",
                    "groups": "",
                    "expiry time": "2017-08-16 08:57:46",
                    "defaults variables": None,
                }
            ]
        ),
    ):
        ret = postgres.user_list(
            "test_user",
            host="test_host",
            port="test_port",
            maintenance_db="maint_db",
            password="test_password",
            runas="foo",
        )

        assert ret == {
            "test_user": {
                "superuser": True,
                "defaults variables": None,
                "can create databases": True,
                "can create roles": True,
                "connections": None,
                "replication": None,
                "expiry time": datetime.datetime(2017, 8, 16, 8, 57, 46),
                "can login": True,
                "can update system catalogs": True,
                "groups": [],
                "inherits privileges": True,
            }
        }


def test_user_remove():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")), patch(
        "salt.modules.postgres.version", Mock(return_value="9.1")
    ), patch(
        "salt.modules.postgres.user_exists", Mock(return_value=True)
    ):
        postgres.user_remove(
            "testuser",
            user="testuser",
            host="testhost",
            port="testport",
            maintenance_db="maint_db",
            password="testpassword",
            runas="foo",
        )
        postgres._run_psql.assert_called_once_with(
            [
                "/usr/bin/pgsql",
                "--no-align",
                "--no-readline",
                "--no-psqlrc",
                "--no-password",
                "--username",
                "testuser",
                "--host",
                "testhost",
                "--port",
                "testport",
                "--dbname",
                "maint_db",
                "-c",
                'DROP ROLE "testuser"',
            ],
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
            runas="foo",
        )


def test_user_update():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch(
        "salt.modules.postgres.role_get",
        Mock(return_value={"superuser": False}),
    ):
        postgres.user_update(
            "test_username",
            user="test_user",
            host="test_host",
            port="test_port",
            maintenance_db="test_maint",
            password="test_pass",
            createdb=False,
            createroles=False,
            encrypted=False,
            inherit=True,
            login=True,
            replication=False,
            rolepassword="test_role_pass",
            valid_until="2017-07-01",
            groups="test_groups",
            runas="foo",
        )
        # postgres._run_psql.call_args[0][0] will contain the list of CLI args.
        # The first 14 elements of this list are initial args used in all (or
        # virtually all) commands run through _run_psql(), so the actual SQL
        # query will be in the 15th argument.
        assert re.match(
            'ALTER ROLE "test_username" WITH  INHERIT NOCREATEDB '
            "NOCREATEROLE NOREPLICATION LOGIN "
            "UNENCRYPTED PASSWORD ['\"]{0,5}test_role_pass['\"]{0,5} "
            "VALID UNTIL '2017-07-01';"
            ' GRANT "test_groups" TO "test_username"',
            postgres._run_psql.call_args[0][0][14],
        )


def test_user_update2():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch(
        "salt.modules.postgres.role_get",
        Mock(return_value={"superuser": False}),
    ):
        postgres.user_update(
            "test_username",
            user="test_user",
            host="test_host",
            port="test_port",
            maintenance_db="test_maint",
            password="test_pass",
            createdb=False,
            createroles=True,
            encrypted=False,
            inherit=True,
            login=True,
            replication=False,
            groups="test_groups",
            runas="foo",
        )
        # postgres._run_psql.call_args[0][0] will contain the list of CLI args.
        # The first 14 elements of this list are initial args used in all (or
        # virtually all) commands run through _run_psql(), so the actual SQL
        # query will be in the 15th argument.
        assert re.match(
            'ALTER ROLE "test_username" WITH  INHERIT NOCREATEDB '
            "CREATEROLE NOREPLICATION LOGIN;"
            ' GRANT "test_groups" TO "test_username"',
            postgres._run_psql.call_args[0][0][14],
        )


def test_user_update3():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch(
        "salt.modules.postgres.role_get",
        Mock(return_value={"superuser": False}),
    ):
        postgres.user_update(
            "test_username",
            user="test_user",
            host="test_host",
            port="test_port",
            maintenance_db="test_maint",
            password="test_pass",
            createdb=False,
            createroles=True,
            encrypted=False,
            inherit=True,
            login=True,
            rolepassword=False,
            replication=False,
            groups="test_groups",
            runas="foo",
        )
        # postgres._run_psql.call_args[0][0] will contain the list of CLI args.
        # The first 14 elements of this list are initial args used in all (or
        # virtually all) commands run through _run_psql(), so the actual SQL
        # query will be in the 15th argument.
        assert re.match(
            'ALTER ROLE "test_username" WITH  INHERIT NOCREATEDB '
            "CREATEROLE NOREPLICATION LOGIN NOPASSWORD;"
            ' GRANT "test_groups" TO "test_username"',
            postgres._run_psql.call_args[0][0][14],
        )


@pytest.mark.skip_on_fips_enabled_platform
def test_user_update_encrypted_passwd():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch(
        "salt.modules.postgres.role_get",
        Mock(return_value={"superuser": False}),
    ):
        postgres.user_update(
            "test_username",
            user="test_user",
            host="test_host",
            port="test_port",
            maintenance_db="test_maint",
            password="test_pass",
            createdb=False,
            createroles=True,
            encrypted=True,
            inherit=True,
            login=True,
            rolepassword="foobar",
            replication=False,
            groups="test_groups",
            runas="foo",
        )
        # postgres._run_psql.call_args[0][0] will contain the list of CLI args.
        # The first 14 elements of this list are initial args used in all (or
        # virtually all) commands run through _run_psql(), so the actual SQL
        # query will be in the 15th argument.
        assert re.match(
            'ALTER ROLE "test_username" WITH  INHERIT NOCREATEDB '
            "CREATEROLE NOREPLICATION LOGIN "
            "ENCRYPTED PASSWORD "
            "['\"]{0,5}md531c27e68d3771c392b52102c01be1da1['\"]{0,5}"
            '; GRANT "test_groups" TO "test_username"',
            postgres._run_psql.call_args[0][0][14],
        )


def test_version():
    with patch(
        "salt.modules.postgres._run_psql",
        Mock(return_value={"retcode": 0, "stdout": "9.1.9"}),
    ):
        postgres.version(
            user="test_user",
            host="test_host",
            port="test_port",
            maintenance_db="test_maint",
            password="test_pass",
            runas="foo",
        )
        # postgres._run_psql.call_args[0][0] will contain the list of CLI args.
        # The first 14 elements of this list are initial args used in all (or
        # virtually all) commands run through _run_psql(), so the actual SQL
        # query will be in the 15th argument.
        assert re.match(
            "SELECT setting FROM pg_catalog.pg_settings",
            postgres._run_psql.call_args[0][0][14],
        )


def test_installed_extensions():
    with patch(
        "salt.modules.postgres.psql_query",
        Mock(return_value=[{"extname": "foo", "extversion": "1"}]),
    ):
        exts = postgres.installed_extensions()
        assert exts == {"foo": {"extversion": "1", "extname": "foo"}}


def test_available_extensions():
    with patch(
        "salt.modules.postgres.psql_query",
        Mock(return_value=[{"name": "foo", "default_version": "1"}]),
    ):
        exts = postgres.available_extensions()
        assert exts == {"foo": {"default_version": "1", "name": "foo"}}


def test_drop_extension2():
    with patch(
        "salt.modules.postgres.installed_extensions", Mock(side_effect=[{}, {}])
    ):
        with patch(
            "salt.modules.postgres._psql_prepare_and_run", Mock(return_value=None)
        ):
            with patch(
                "salt.modules.postgres.available_extensions",
                Mock(return_value={"foo": {"default_version": "1", "name": "foo"}}),
            ):
                assert postgres.drop_extension("foo")


def test_drop_extension3():
    with patch(
        "salt.modules.postgres.installed_extensions",
        Mock(side_effect=[{"foo": {"extversion": "1", "extname": "foo"}}, {}]),
    ):
        with patch(
            "salt.modules.postgres._psql_prepare_and_run", Mock(return_value=None)
        ):
            with patch(
                "salt.modules.postgres.available_extensions",
                Mock(return_value={"foo": {"default_version": "1", "name": "foo"}}),
            ):
                assert postgres.drop_extension("foo")


def test_drop_extension1():
    with patch(
        "salt.modules.postgres.installed_extensions",
        Mock(
            side_effect=[
                {"foo": {"extversion": "1", "extname": "foo"}},
                {"foo": {"extversion": "1", "extname": "foo"}},
            ]
        ),
    ):
        with patch(
            "salt.modules.postgres._psql_prepare_and_run", Mock(return_value=None)
        ):
            with patch(
                "salt.modules.postgres.available_extensions",
                Mock(return_value={"foo": {"default_version": "1", "name": "foo"}}),
            ):
                assert not postgres.drop_extension("foo")


def test_create_mtdata():
    with patch(
        "salt.modules.postgres.installed_extensions",
        Mock(
            return_value={
                "foo": {
                    "extversion": "0.8",
                    "extrelocatable": "t",
                    "schema_name": "foo",
                    "extname": "foo",
                }
            },
        ),
    ):
        with patch(
            "salt.modules.postgres.available_extensions",
            Mock(return_value={"foo": {"default_version": "1.4", "name": "foo"}}),
        ):
            ret = postgres.create_metadata("foo", schema="bar", ext_version="1.4")
            assert postgres._EXTENSION_INSTALLED in ret
            assert postgres._EXTENSION_TO_UPGRADE in ret
            assert postgres._EXTENSION_TO_MOVE in ret

            ret = postgres.create_metadata("foo", schema="foo", ext_version="0.4")
            assert postgres._EXTENSION_INSTALLED in ret
            assert postgres._EXTENSION_TO_UPGRADE not in ret
            assert postgres._EXTENSION_TO_MOVE not in ret

            ret = postgres.create_metadata("foo")
            assert postgres._EXTENSION_INSTALLED in ret
            assert postgres._EXTENSION_TO_UPGRADE not in ret
            assert postgres._EXTENSION_TO_MOVE not in ret

            ret = postgres.create_metadata("foobar")
            assert postgres._EXTENSION_NOT_INSTALLED in ret
            assert postgres._EXTENSION_INSTALLED not in ret
            assert postgres._EXTENSION_TO_UPGRADE not in ret
            assert postgres._EXTENSION_TO_MOVE not in ret


def test_create_extension_newerthan():
    """
    scenario of creating upgrading extensions with possible schema and
    version specifications
    """
    with patch(
        "salt.modules.postgres.create_metadata",
        Mock(
            side_effect=[
                # create succeeded
                [postgres._EXTENSION_NOT_INSTALLED],
                [postgres._EXTENSION_INSTALLED],
                [postgres._EXTENSION_NOT_INSTALLED],
                [postgres._EXTENSION_INSTALLED],
                # create failed
                [postgres._EXTENSION_NOT_INSTALLED],
                [postgres._EXTENSION_NOT_INSTALLED],
                # move+upgrade succeeded
                [
                    postgres._EXTENSION_TO_MOVE,
                    postgres._EXTENSION_TO_UPGRADE,
                    postgres._EXTENSION_INSTALLED,
                ],
                [postgres._EXTENSION_INSTALLED],
                # move succeeded
                [postgres._EXTENSION_TO_MOVE, postgres._EXTENSION_INSTALLED],
                [postgres._EXTENSION_INSTALLED],
                # upgrade succeeded
                [postgres._EXTENSION_TO_UPGRADE, postgres._EXTENSION_INSTALLED],
                [postgres._EXTENSION_INSTALLED],
                # upgrade failed
                [postgres._EXTENSION_TO_UPGRADE, postgres._EXTENSION_INSTALLED],
                [postgres._EXTENSION_TO_UPGRADE, postgres._EXTENSION_INSTALLED],
                # move failed
                [postgres._EXTENSION_TO_MOVE, postgres._EXTENSION_INSTALLED],
                [postgres._EXTENSION_TO_MOVE, postgres._EXTENSION_INSTALLED],
            ]
        ),
    ):
        with patch(
            "salt.modules.postgres._psql_prepare_and_run", Mock(return_value=None)
        ):
            with patch(
                "salt.modules.postgres.available_extensions",
                Mock(return_value={"foo": {"default_version": "1.4", "name": "foo"}}),
            ):
                assert postgres.create_extension("foo")
                assert re.match(
                    'CREATE EXTENSION IF NOT EXISTS "foo" ;',
                    postgres._psql_prepare_and_run.call_args[0][0][1],
                )

                assert postgres.create_extension(
                    "foo", schema="a", ext_version="b", from_version="c"
                )
                assert re.match(
                    'CREATE EXTENSION IF NOT EXISTS "foo" '
                    'WITH SCHEMA "a" VERSION b FROM c ;',
                    postgres._psql_prepare_and_run.call_args[0][0][1],
                )
                assert not postgres.create_extension("foo")

                ret = postgres.create_extension("foo", ext_version="a", schema="b")
                assert ret is True
                assert re.match(
                    'ALTER EXTENSION "foo" SET SCHEMA "b";'
                    ' ALTER EXTENSION "foo" UPDATE TO a;',
                    postgres._psql_prepare_and_run.call_args[0][0][1],
                )

                ret = postgres.create_extension("foo", ext_version="a", schema="b")
                assert ret is True
                assert re.match(
                    'ALTER EXTENSION "foo" SET SCHEMA "b";',
                    postgres._psql_prepare_and_run.call_args[0][0][1],
                )

                ret = postgres.create_extension("foo", ext_version="a", schema="b")
                assert ret is True
                assert re.match(
                    'ALTER EXTENSION "foo" UPDATE TO a;',
                    postgres._psql_prepare_and_run.call_args[0][0][1],
                )
                assert not postgres.create_extension("foo", ext_version="a", schema="b")
                assert not postgres.create_extension("foo", ext_version="a", schema="b")


@pytest.mark.skip_on_fips_enabled_platform
def test_encrypt_passwords():
    assert postgres._maybe_encrypt_password("foo", "bar", False) == "bar"
    assert (
        postgres._maybe_encrypt_password("foo", "bar", True)
        == "md596948aad3fcae80c08a35c9b5958cd89"
    )


def test_schema_list(get_test_list_schema_csv):
    with patch(
        "salt.modules.postgres._run_psql",
        Mock(return_value={"retcode": 0, "stdout": get_test_list_schema_csv}),
    ):
        ret = postgres.schema_list(
            "maint_db",
            db_user="testuser",
            db_host="testhost",
            db_port="testport",
            db_password="foo",
        )
        assert ret == {
            "public": {
                "acl": "{postgres=UC/postgres,=UC/postgres}",
                "owner": "postgres",
            },
            "pg_toast": {"acl": "", "owner": "postgres"},
        }


def test_schema_exists():
    with patch("salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})):
        with patch(
            "salt.modules.postgres.psql_query",
            Mock(
                return_value=[
                    {
                        "name": "public",
                        "acl": "{postgres=UC/postgres,=UC/postgres}",
                        "owner": "postgres",
                    }
                ]
            ),
        ):
            ret = postgres.schema_exists("template1", "public")
            assert ret is True


def test_schema_get():
    with patch("salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})):
        with patch(
            "salt.modules.postgres.psql_query",
            Mock(
                return_value=[
                    {
                        "name": "public",
                        "acl": "{postgres=UC/postgres,=UC/postgres}",
                        "owner": "postgres",
                    }
                ]
            ),
        ):
            ret = postgres.schema_get("template1", "public")
            assert ret == {
                "acl": "{postgres=UC/postgres,=UC/postgres}",
                "owner": "postgres",
            }


def test_schema_get_again():
    with patch("salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})):
        with patch(
            "salt.modules.postgres.psql_query",
            Mock(
                return_value=[
                    {
                        "name": "public",
                        "acl": "{postgres=UC/postgres,=UC/postgres}",
                        "owner": "postgres",
                    }
                ]
            ),
        ):
            ret = postgres.schema_get("template1", "pg_toast")
            assert ret is None


def test_schema_create():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        with patch("salt.modules.postgres.schema_exists", Mock(return_value=False)):
            postgres.schema_create(
                "maint_db",
                "testschema",
                user="user",
                db_host="testhost",
                db_port="testport",
                db_user="testuser",
                db_password="testpassword",
            )
            postgres._run_psql.assert_called_once_with(
                [
                    "/usr/bin/pgsql",
                    "--no-align",
                    "--no-readline",
                    "--no-psqlrc",
                    "--no-password",
                    "--username",
                    "testuser",
                    "--host",
                    "testhost",
                    "--port",
                    "testport",
                    "--dbname",
                    "maint_db",
                    "-c",
                    'CREATE SCHEMA "testschema"',
                ],
                host="testhost",
                port="testport",
                password="testpassword",
                user="testuser",
                runas="user",
            )


def test_schema_create2():
    with patch("salt.modules.postgres.schema_exists", Mock(return_value=True)):
        ret = postgres.schema_create(
            "test_db",
            "test_schema",
            user="user",
            db_host="test_host",
            db_port="test_port",
            db_user="test_user",
            db_password="test_password",
        )
        assert ret is False


def test_schema_remove():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        with patch("salt.modules.postgres.schema_exists", Mock(return_value=True)):
            postgres.schema_remove(
                "maint_db",
                "testschema",
                user="user",
                db_host="testhost",
                db_port="testport",
                db_user="testuser",
                db_password="testpassword",
            )
            postgres._run_psql.assert_called_once_with(
                [
                    "/usr/bin/pgsql",
                    "--no-align",
                    "--no-readline",
                    "--no-psqlrc",
                    "--no-password",
                    "--username",
                    "testuser",
                    "--host",
                    "testhost",
                    "--port",
                    "testport",
                    "--dbname",
                    "maint_db",
                    "-c",
                    'DROP SCHEMA "testschema"',
                ],
                host="testhost",
                port="testport",
                password="testpassword",
                user="testuser",
                runas="user",
            )


def test_schema_remove2():
    with patch("salt.modules.postgres.schema_exists", Mock(return_value=False)):
        ret = postgres.schema_remove(
            "test_db",
            "test_schema",
            user="user",
            db_host="test_host",
            db_port="test_port",
            db_user="test_user",
            db_password="test_password",
        )
        assert ret is False


def test_language_list(get_test_list_language_csv):
    """
    Test language listing
    """
    with patch(
        "salt.modules.postgres._run_psql",
        Mock(return_value={"retcode": 0, "stdout": get_test_list_language_csv}),
    ):
        ret = postgres.language_list(
            "testdb",
            user="testuser",
            host="testhost",
            port="testport",
            password="foo",
        )
        assert ret == {
            "c": "c",
            "internal": "internal",
            "plpgsql": "plpgsql",
            "sql": "sql",
        }


def test_language_exists():
    """
    Test language existence check
    """
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch(
        "salt.modules.postgres.psql_query",
        Mock(
            return_value=[
                {"Name": "internal"},
                {"Name": "c"},
                {"Name": "sql"},
                {"Name": "plpgsql"},
            ]
        ),
    ), patch(
        "salt.modules.postgres.language_exists", Mock(return_value=True)
    ):
        ret = postgres.language_exists("sql", "testdb")
        assert ret is True


def test_language_create():
    """
    Test language creation - does not exist in db
    """
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        with patch("salt.modules.postgres.language_exists", Mock(return_value=False)):
            postgres.language_create(
                "plpythonu",
                "testdb",
                runas="user",
                host="testhost",
                port="testport",
                user="testuser",
                password="testpassword",
            )
            postgres._run_psql.assert_called_once_with(
                [
                    "/usr/bin/pgsql",
                    "--no-align",
                    "--no-readline",
                    "--no-psqlrc",
                    "--no-password",
                    "--username",
                    "testuser",
                    "--host",
                    "testhost",
                    "--port",
                    "testport",
                    "--dbname",
                    "testdb",
                    "-c",
                    "CREATE LANGUAGE plpythonu",
                ],
                host="testhost",
                port="testport",
                password="testpassword",
                user="testuser",
                runas="user",
            )


def test_language_create_exists():
    """
    Test language creation - already exists in db
    """
    with patch("salt.modules.postgres.language_exists", Mock(return_value=True)):
        ret = postgres.language_create(
            "plpythonu",
            "testdb",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )
        assert ret is False


def test_language_remove():
    """
    Test language removal - exists in db
    """
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        with patch("salt.modules.postgres.language_exists", Mock(return_value=True)):
            postgres.language_remove(
                "plpgsql",
                "testdb",
                runas="user",
                host="testhost",
                port="testport",
                user="testuser",
                password="testpassword",
            )
            postgres._run_psql.assert_called_once_with(
                [
                    "/usr/bin/pgsql",
                    "--no-align",
                    "--no-readline",
                    "--no-psqlrc",
                    "--no-password",
                    "--username",
                    "testuser",
                    "--host",
                    "testhost",
                    "--port",
                    "testport",
                    "--dbname",
                    "testdb",
                    "-c",
                    "DROP LANGUAGE plpgsql",
                ],
                host="testhost",
                port="testport",
                password="testpassword",
                user="testuser",
                runas="user",
            )


def test_language_remove_non_exist():
    """
    Test language removal - does not exist in db
    """
    with patch("salt.modules.postgres.language_exists", Mock(return_value=False)):
        ret = postgres.language_remove(
            "plpgsql",
            "testdb",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )
        assert ret is False


def test_privileges_list_table(get_test_privileges_list_table_csv):
    """
    Test privilege listing on a table
    """
    with patch(
        "salt.modules.postgres._run_psql",
        Mock(return_value={"retcode": 0, "stdout": get_test_privileges_list_table_csv}),
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        ret = postgres.privileges_list(
            "awl",
            "table",
            maintenance_db="db_name",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )
        expected = {
            "bayestest": {
                "INSERT": False,
                "UPDATE": False,
                "SELECT": False,
                "DELETE": False,
            },
            "baruwa": {
                "INSERT": True,
                "TRUNCATE": True,
                "UPDATE": True,
                "TRIGGER": True,
                "REFERENCES": True,
                "SELECT": True,
                "DELETE": True,
            },
            "baruwatest": {
                "INSERT": False,
                "TRUNCATE": False,
                "UPDATE": False,
                "TRIGGER": False,
                "REFERENCES": False,
                "SELECT": False,
                "DELETE": False,
            },
        }
        assert ret == expected

        query = (
            "COPY (SELECT relacl AS name FROM pg_catalog.pg_class c "
            "JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace "
            "WHERE nspname = 'public' AND relname = 'awl' AND relkind in ('r', 'v') "
            "ORDER BY relname) TO STDOUT WITH CSV HEADER"
        )

        postgres._run_psql.assert_called_once_with(
            [
                "/usr/bin/pgsql",
                "--no-align",
                "--no-readline",
                "--no-psqlrc",
                "--no-password",
                "--username",
                "testuser",
                "--host",
                "testhost",
                "--port",
                "testport",
                "--dbname",
                "db_name",
                "-v",
                "datestyle=ISO,MDY",
                "-c",
                query,
            ],
            host="testhost",
            port="testport",
            password="testpassword",
            user="testuser",
            runas="user",
        )


def test_privileges_list_group(get_test_privileges_list_group_csv):
    """
    Test privilege listing on a group
    """
    with patch(
        "salt.modules.postgres._run_psql",
        Mock(return_value={"retcode": 0, "stdout": get_test_privileges_list_group_csv}),
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        ret = postgres.privileges_list(
            "admin",
            "group",
            maintenance_db="db_name",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )
        expected = {
            "baruwa": False,
            "baruwatest": False,
            "baruwatest2": True,
        }
        assert ret == expected

        query = (
            "COPY (SELECT rolname, admin_option "
            "FROM pg_catalog.pg_auth_members m JOIN pg_catalog.pg_roles r "
            "ON m.member=r.oid WHERE m.roleid IN (SELECT oid FROM "
            "pg_catalog.pg_roles WHERE rolname='admin') ORDER BY rolname) "
            "TO STDOUT WITH CSV HEADER"
        )

        postgres._run_psql.assert_called_once_with(
            [
                "/usr/bin/pgsql",
                "--no-align",
                "--no-readline",
                "--no-psqlrc",
                "--no-password",
                "--username",
                "testuser",
                "--host",
                "testhost",
                "--port",
                "testport",
                "--dbname",
                "db_name",
                "-v",
                "datestyle=ISO,MDY",
                "-c",
                query,
            ],
            host="testhost",
            port="testport",
            password="testpassword",
            user="testuser",
            runas="user",
        )


def test_has_privileges_on_table(get_test_privileges_list_table_csv):
    """
    Test privilege checks on table
    """
    with patch(
        "salt.modules.postgres._run_psql",
        Mock(return_value={"retcode": 0, "stdout": get_test_privileges_list_table_csv}),
    ):
        ret = postgres.has_privileges(
            "baruwa",
            "awl",
            "table",
            "SELECT,INSERT",
            grant_option=True,
            maintenance_db="db_name",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )
        assert ret is True

        ret = postgres.has_privileges(
            "baruwa",
            "awl",
            "table",
            "ALL",
            grant_option=True,
            maintenance_db="db_name",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )
        assert ret is True

        ret = postgres.has_privileges(
            "baruwa",
            "awl",
            "table",
            "ALL",
            grant_option=False,
            maintenance_db="db_name",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )
        assert ret is True

        ret = postgres.has_privileges(
            "bayestest",
            "awl",
            "table",
            "SELECT,INSERT,TRUNCATE",
            maintenance_db="db_name",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )
        assert ret is False

        ret = postgres.has_privileges(
            "bayestest",
            "awl",
            "table",
            "SELECT,INSERT",
            maintenance_db="db_name",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )
        assert ret is True


def test_has_privileges_on_group(get_test_privileges_list_group_csv):
    """
    Test privilege checks on group
    """
    with patch(
        "salt.modules.postgres._run_psql",
        Mock(return_value={"retcode": 0, "stdout": get_test_privileges_list_group_csv}),
    ):
        ret = postgres.has_privileges(
            "baruwa",
            "admin",
            "group",
            maintenance_db="db_name",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )
        assert ret is True

        ret = postgres.has_privileges(
            "baruwa",
            "admin",
            "group",
            grant_option=True,
            maintenance_db="db_name",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )
        assert ret is False

        ret = postgres.has_privileges(
            "tony",
            "admin",
            "group",
            maintenance_db="db_name",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )
        assert ret is False


def test_privileges_grant_table():
    """
    Test granting privileges on table
    """
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        with patch("salt.modules.postgres.has_privileges", Mock(return_value=False)):
            ret = postgres.privileges_grant(
                "baruwa",
                "awl",
                "table",
                "ALL",
                grant_option=True,
                maintenance_db="db_name",
                runas="user",
                host="testhost",
                port="testport",
                user="testuser",
                password="testpassword",
            )

            query = 'GRANT ALL ON TABLE public."awl" TO "baruwa" WITH GRANT OPTION'

            postgres._run_psql.assert_called_once_with(
                [
                    "/usr/bin/pgsql",
                    "--no-align",
                    "--no-readline",
                    "--no-psqlrc",
                    "--no-password",
                    "--username",
                    "testuser",
                    "--host",
                    "testhost",
                    "--port",
                    "testport",
                    "--dbname",
                    "db_name",
                    "-c",
                    query,
                ],
                host="testhost",
                port="testport",
                password="testpassword",
                user="testuser",
                runas="user",
            )

    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")), patch(
        "salt.modules.postgres.has_privileges", Mock(return_value=False)
    ):
        ret = postgres.privileges_grant(
            "baruwa",
            "awl",
            "table",
            "ALL",
            maintenance_db="db_name",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )

        query = 'GRANT ALL ON TABLE public."awl" TO "baruwa"'

        postgres._run_psql.assert_called_once_with(
            [
                "/usr/bin/pgsql",
                "--no-align",
                "--no-readline",
                "--no-psqlrc",
                "--no-password",
                "--username",
                "testuser",
                "--host",
                "testhost",
                "--port",
                "testport",
                "--dbname",
                "db_name",
                "-c",
                query,
            ],
            host="testhost",
            port="testport",
            password="testpassword",
            user="testuser",
            runas="user",
        )

    # Test grant on all tables
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")), patch(
        "salt.modules.postgres.has_privileges", Mock(return_value=False)
    ):
        ret = postgres.privileges_grant(
            "baruwa",
            "ALL",
            "table",
            "SELECT",
            maintenance_db="db_name",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )

        query = 'GRANT SELECT ON ALL TABLES IN SCHEMA public TO "baruwa"'

        postgres._run_psql.assert_called_once_with(
            [
                "/usr/bin/pgsql",
                "--no-align",
                "--no-readline",
                "--no-psqlrc",
                "--no-password",
                "--username",
                "testuser",
                "--host",
                "testhost",
                "--port",
                "testport",
                "--dbname",
                "db_name",
                "-c",
                query,
            ],
            host="testhost",
            port="testport",
            password="testpassword",
            user="testuser",
            runas="user",
        )


def test_privileges_grant_group():
    """
    Test granting privileges on group
    """
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")), patch(
        "salt.modules.postgres.has_privileges", Mock(return_value=False)
    ):
        ret = postgres.privileges_grant(
            "baruwa",
            "admins",
            "group",
            grant_option=True,
            maintenance_db="db_name",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )

        query = 'GRANT admins TO "baruwa" WITH ADMIN OPTION'

        postgres._run_psql.assert_called_once_with(
            [
                "/usr/bin/pgsql",
                "--no-align",
                "--no-readline",
                "--no-psqlrc",
                "--no-password",
                "--username",
                "testuser",
                "--host",
                "testhost",
                "--port",
                "testport",
                "--dbname",
                "db_name",
                "-c",
                query,
            ],
            host="testhost",
            port="testport",
            password="testpassword",
            user="testuser",
            runas="user",
        )

    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")), patch(
        "salt.modules.postgres.has_privileges", Mock(return_value=False)
    ):
        ret = postgres.privileges_grant(
            "baruwa",
            "admins",
            "group",
            maintenance_db="db_name",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )

        query = 'GRANT admins TO "baruwa"'

        postgres._run_psql.assert_called_once_with(
            [
                "/usr/bin/pgsql",
                "--no-align",
                "--no-readline",
                "--no-psqlrc",
                "--no-password",
                "--username",
                "testuser",
                "--host",
                "testhost",
                "--port",
                "testport",
                "--dbname",
                "db_name",
                "-c",
                query,
            ],
            host="testhost",
            port="testport",
            password="testpassword",
            user="testuser",
            runas="user",
        )


def test_privileges_revoke_table():
    """
    Test revoking privileges on table
    """
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")), patch(
        "salt.modules.postgres.has_privileges", Mock(return_value=True)
    ):
        ret = postgres.privileges_revoke(
            "baruwa",
            "awl",
            "table",
            "ALL",
            maintenance_db="db_name",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )

        query = "REVOKE ALL ON TABLE public.awl FROM baruwa"

        postgres._run_psql.assert_called_once_with(
            [
                "/usr/bin/pgsql",
                "--no-align",
                "--no-readline",
                "--no-psqlrc",
                "--no-password",
                "--username",
                "testuser",
                "--host",
                "testhost",
                "--port",
                "testport",
                "--dbname",
                "db_name",
                "-c",
                query,
            ],
            host="testhost",
            port="testport",
            password="testpassword",
            user="testuser",
            runas="user",
        )


def test_privileges_revoke_group():
    """
    Test revoking privileges on group
    """
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")), patch(
        "salt.modules.postgres.has_privileges", Mock(return_value=True)
    ):
        ret = postgres.privileges_revoke(
            "baruwa",
            "admins",
            "group",
            maintenance_db="db_name",
            runas="user",
            host="testhost",
            port="testport",
            user="testuser",
            password="testpassword",
        )

        query = "REVOKE admins FROM baruwa"

        postgres._run_psql.assert_called_once_with(
            [
                "/usr/bin/pgsql",
                "--no-align",
                "--no-readline",
                "--no-psqlrc",
                "--no-password",
                "--username",
                "testuser",
                "--host",
                "testhost",
                "--port",
                "testport",
                "--dbname",
                "db_name",
                "-c",
                query,
            ],
            host="testhost",
            port="testport",
            password="testpassword",
            user="testuser",
            runas="user",
        )


def test_datadir_init():
    """
    Test Initializing a postgres data directory
    """
    with patch("salt.modules.postgres._run_initdb", Mock(return_value={"retcode": 0})):
        with patch("salt.modules.postgres.datadir_exists", Mock(return_value=False)):
            name = "/var/lib/pgsql/data"
            ret = postgres.datadir_init(
                name, user="postgres", password="test", runas="postgres"
            )
            postgres._run_initdb.assert_called_once_with(
                name,
                auth="password",
                encoding="UTF8",
                locale=None,
                password="test",
                runas="postgres",
                checksums=False,
                waldir=None,
                user="postgres",
            )
            assert ret is True


def test_datadir_exists():
    """
    Test Checks if postgres data directory has been initialized
    """
    with patch("os.path.isfile", Mock(return_value=True)):
        name = "/var/lib/pgsql/data"
        ret = postgres.datadir_exists(name)
        assert ret is True


@pytest.mark.parametrize(
    "v1,v2,result",
    (
        ("8.5", "9.5", True),
        ("8.5", "8.6", True),
        ("8.5.2", "8.5.3", True),
        ("9.5", "8.5", False),
        ("9.5", "9.6", True),
        ("9.5.0", "9.5.1", True),
        ("9.5", "9.5.1", True),
        ("9.5.1", "9.5", False),
        ("9.5b", "9.5a", False),
        ("10a", "10b", True),
        ("1.2.3.4", "1.2.3.5", True),
        ("10dev", "10next", True),
        ("10next", "10dev", False),
    ),
)
def test_pg_is_older_ext_ver(v1, v2, result):
    """
    Test Checks if postgres extension version string is older
    """
    assert postgres._pg_is_older_ext_ver(v1, v2) is result


def test_tablespace_create():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        postgres.tablespace_create(
            "test_tablespace",
            "/tmp/postgres_test_tablespace",
            user="testuser",
            host="testhost",
            port="testport",
            maintenance_db="maint_db",
            password="foo",
            owner="otheruser",
            runas="foo",
        )

        postgres._run_psql.assert_called_once_with(
            [
                "/usr/bin/pgsql",
                "--no-align",
                "--no-readline",
                "--no-psqlrc",
                "--no-password",
                "--username",
                "testuser",
                "--host",
                "testhost",
                "--port",
                "testport",
                "--dbname",
                "maint_db",
                "-c",
                'CREATE TABLESPACE "test_tablespace" OWNER "otheruser" LOCATION \'/tmp/postgres_test_tablespace\' ',
            ],
            runas="foo",
            password="foo",
            host="testhost",
            port="testport",
            user="testuser",
        )


def test_tablespace_list():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch(
        "salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")
    ), patch.dict(
        postgres.__salt__,
        {
            "postgres.psql_query": MagicMock(
                return_value=[
                    {
                        "Name": "pg_global",
                        "Owner": "postgres",
                        "ACL": "",
                        "Opts": "",
                        "Location": "",
                    },
                    {
                        "Name": "pg_default",
                        "Owner": "postgres",
                        "ACL": "",
                        "Opts": "",
                        "Location": "",
                    },
                    {
                        "Name": "test_tablespace",
                        "Owner": "testuser",
                        "ACL": "",
                        "Opts": "",
                        "Location": "/tmp/posgrest_test_tablespace",
                    },
                ]
            ),
        },
    ):
        ret = postgres.tablespace_list(
            user="testuser",
            host="testhost",
            port="testport",
            maintenance_db="maint_db",
            password="foo",
            runas="foo",
        )

        expected_data = {
            "pg_global": {"Owner": "postgres", "ACL": "", "Opts": "", "Location": ""},
            "pg_default": {"Owner": "postgres", "ACL": "", "Opts": "", "Location": ""},
            "test_tablespace": {
                "Owner": "testuser",
                "ACL": "",
                "Opts": "",
                "Location": "/tmp/posgrest_test_tablespace",
            },
        }
        assert ret == expected_data


def test_tablespace_exists_true():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch(
        "salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")
    ), patch.dict(
        postgres.__salt__,
        {
            "postgres.psql_query": MagicMock(
                return_value=[
                    {
                        "Name": "pg_global",
                        "Owner": "postgres",
                        "ACL": "",
                        "Opts": "",
                        "Location": "",
                    },
                    {
                        "Name": "pg_default",
                        "Owner": "postgres",
                        "ACL": "",
                        "Opts": "",
                        "Location": "",
                    },
                    {
                        "Name": "test_tablespace",
                        "Owner": "testuser",
                        "ACL": "",
                        "Opts": "",
                        "Location": "/tmp/posgrest_test_tablespace",
                    },
                ]
            ),
        },
    ):
        ret = postgres.tablespace_exists(
            "test_tablespace",
            user="testuser",
            host="testhost",
            port="testport",
            maintenance_db="maint_db",
            password="foo",
            runas="foo",
        )
        assert ret is True


def test_tablespace_exists_false():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch(
        "salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")
    ), patch.dict(
        postgres.__salt__,
        {
            "postgres.psql_query": MagicMock(
                return_value=[
                    {
                        "Name": "pg_global",
                        "Owner": "postgres",
                        "ACL": "",
                        "Opts": "",
                        "Location": "",
                    },
                    {
                        "Name": "pg_default",
                        "Owner": "postgres",
                        "ACL": "",
                        "Opts": "",
                        "Location": "",
                    },
                    {
                        "Name": "test_tablespace",
                        "Owner": "testuser",
                        "ACL": "",
                        "Opts": "",
                        "Location": "/tmp/posgrest_test_tablespace",
                    },
                ]
            ),
        },
    ):
        ret = postgres.tablespace_exists(
            "bad_test_tablespace",
            user="testuser",
            host="testhost",
            port="testport",
            maintenance_db="maint_db",
            password="foo",
            runas="foo",
        )
        assert ret is False


def test_tablespace_alter_new_owner():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        postgres.tablespace_alter(
            "test_tablespace",
            user="testuser",
            host="testhost",
            port="testport",
            maintenance_db="maint_db",
            password="foo",
            runas="foo",
            new_owner="testuser",
        )

        postgres._run_psql.assert_called_once_with(
            [
                "/usr/bin/pgsql",
                "--no-align",
                "--no-readline",
                "--no-psqlrc",
                "--no-password",
                "--username",
                "testuser",
                "--host",
                "testhost",
                "--port",
                "testport",
                "--dbname",
                "maint_db",
                "-c",
                'ALTER TABLESPACE "test_tablespace" OWNER TO "testuser"',
            ],
            runas="foo",
            password="foo",
            host="testhost",
            port="testport",
            user="testuser",
        )


def test_tablespace_alter_new_name():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        postgres.tablespace_alter(
            "test_tablespace",
            user="testuser",
            host="testhost",
            port="testport",
            maintenance_db="maint_db",
            password="foo",
            runas="foo",
            new_name="test_tablespace2",
        )

        postgres._run_psql.assert_called_once_with(
            [
                "/usr/bin/pgsql",
                "--no-align",
                "--no-readline",
                "--no-psqlrc",
                "--no-password",
                "--username",
                "testuser",
                "--host",
                "testhost",
                "--port",
                "testport",
                "--dbname",
                "maint_db",
                "-c",
                'ALTER TABLESPACE "test_tablespace" RENAME TO "test_tablespace2"',
            ],
            runas="foo",
            password="foo",
            host="testhost",
            port="testport",
            user="testuser",
        )


def test_tablespace_remove():
    with patch(
        "salt.modules.postgres._run_psql", Mock(return_value={"retcode": 0})
    ), patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/pgsql")):
        postgres.tablespace_remove(
            "test_tablespace",
            user="testuser",
            host="testhost",
            port="testport",
            maintenance_db="maint_db",
            password="foo",
            runas="foo",
        )

        postgres._run_psql.assert_called_once_with(
            [
                "/usr/bin/pgsql",
                "--no-align",
                "--no-readline",
                "--no-psqlrc",
                "--no-password",
                "--username",
                "testuser",
                "--host",
                "testhost",
                "--port",
                "testport",
                "--dbname",
                "maint_db",
                "-c",
                'DROP TABLESPACE "test_tablespace"',
            ],
            runas="foo",
            password="foo",
            host="testhost",
            port="testport",
            user="testuser",
        )



#
# Copyright 2016 SUSE LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
    :codeauthor: Bo Maryniuk <bo@suse.de>
"""

import io
from collections import OrderedDict

from salt.modules.inspectlib.entities import CsvDBEntity
from salt.modules.inspectlib.fsdb import CsvDB
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


def mock_open(data=None):
    """
    Mock "open" function in a simple way.

    :param data:
    :return:
    """
    data = io.StringIO(data)
    mock = MagicMock(spec=io.FileIO)
    handle = MagicMock(spec=io.FileIO)
    handle.write.return_value = None
    handle.__enter__.return_value = data or handle
    mock.return_value = handle

    return mock


class Writable(io.StringIO):
    def __init__(self, data=None):
        if data:
            io.StringIO.__init__(self, data)
        else:
            io.StringIO.__init__(self)
        self.data = []

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self

    def __enter__(self):
        return self

    def write(self, s):
        self.data.append(s)


class FoobarEntity(CsvDBEntity):
    """
    Entity for test purposes.
    """

    _TABLE = "some_table"

    def __init__(self):
        self.foo = 0
        self.bar = ""
        self.spam = 0.0


class InspectorFSDBTestCase(TestCase):
    """
    Test case for the FSDB: FileSystem Database.

    FSDB is a very simple object-to-CSV storage with a very inefficient
    update/delete operations (nice to have at some point) and efficient
    storing/reading the objects (what is exactly needed for the functionality).

    Main advantage of FSDB is to store Python objects in just a CSV files,
    and have a very small code base.
    """

    def setUp(self):
        patcher = patch("os.makedirs", MagicMock())
        patcher.start()
        self.addCleanup(patcher.stop)

    def test_open(self):
        """
        Test opening the database.
        :return:
        """
        with patch("os.listdir", MagicMock(return_value=["test_db"])), patch(
            "gzip.open", mock_open("foo:int,bar:str")
        ):
            csvdb = CsvDB("/foobar")
            csvdb.open()
            assert list(csvdb.list_tables()) == ["test_db"]
            assert csvdb.is_closed() is False

    def test_close(self):
        """
        Test closing the database.
        :return:
        """
        with patch("os.listdir", MagicMock(return_value=["test_db"])), patch(
            "gzip.open", mock_open("foo:int,bar:str")
        ):
            csvdb = CsvDB("/foobar")
            csvdb.open()
            csvdb.close()
            assert csvdb.is_closed() is True

    def test_create_table(self):
        """
        Test creating table.
        :return:
        """
        with patch("os.path.exists", MagicMock(return_value=False)), patch(
            "os.listdir", MagicMock(return_value=["some_table"])
        ):
            writable = Writable()
            with patch("gzip.open", MagicMock(return_value=writable)) as gzip_mock_open:
                csvdb = CsvDB("/foobar")
                csvdb.open()
                csvdb.create_table_from_object(FoobarEntity())

            # test the second call to gzip.open, the first is in the list_tables function
            assert gzip_mock_open.call_args_list[1][0][1] == "wt"

            sorted_writable_data = sorted(writable.data[0].strip().split(","))
            sorted_expected_data = sorted("foo:int,bar:str,spam:float".split(","))
            self.assertEqual(sorted_writable_data, sorted_expected_data)

    def test_list_databases(self):
        """
        Test list databases.
        :return:
        """
        with patch("os.listdir", MagicMock(return_value=["test_db"])):
            csvdb = CsvDB("/foobar")
            assert csvdb.list() == ["test_db"]

    def test_add_object(self):
        """
        Test storing object into the database.
        :return:
        """
        with patch("os.path.exists", MagicMock(return_value=False)), patch(
            "os.listdir", MagicMock(return_value=["some_table"])
        ):
            writable = Writable()
            with patch("gzip.open", MagicMock(return_value=writable)) as gzip_mock_open:
                obj = FoobarEntity()
                obj.foo = 123
                obj.bar = "test entity"
                obj.spam = 0.123

                csvdb = CsvDB("/foobar")
                csvdb.open()
                csvdb._tables = {
                    "some_table": OrderedDict(
                        [
                            tuple(elm.split(":"))
                            for elm in ["foo:int", "bar:str", "spam:float"]
                        ]
                    )
                }
                csvdb.store(obj)

                # test the second call to gzip.open, the first is in the list_tables function
                assert gzip_mock_open.call_args_list[1][0][1] == "at"

                assert writable.data[0].strip() == "123,test entity,0.123"

    def test_delete_object(self):
        """
        Deleting an object from the store.
        :return:
        """
        with patch("gzip.open", MagicMock()), patch(
            "os.listdir", MagicMock(return_value=["test_db"])
        ), patch(
            "csv.reader",
            MagicMock(
                return_value=iter(
                    [
                        [],
                        ["foo:int", "bar:str", "spam:float"],
                        ["123", "test", "0.123"],
                        ["234", "another", "0.456"],
                    ]
                )
            ),
        ):

            class InterceptedCsvDB(CsvDB):
                def __init__(self, path):
                    CsvDB.__init__(self, path)
                    self._remained = list()

                def store(self, obj, distinct=False):
                    self._remained.append(obj)

            csvdb = InterceptedCsvDB("/foobar")
            csvdb.open()
            csvdb.create_table_from_object = MagicMock()
            csvdb.flush = MagicMock()

            assert csvdb.delete(FoobarEntity, eq={"foo": 123}) is True
            assert len(csvdb._remained) == 1

            assert csvdb._remained[0].foo == 234
            assert csvdb._remained[0].bar == "another"
            assert csvdb._remained[0].spam == 0.456

    def test_update_object(self):
        """
        Updating an object from the store.
        :return:
        """
        with patch("gzip.open", MagicMock()), patch(
            "os.listdir", MagicMock(return_value=["test_db"])
        ), patch(
            "csv.reader",
            MagicMock(
                return_value=iter(
                    [
                        [],
                        ["foo:int", "bar:str", "spam:float"],
                        ["123", "test", "0.123"],
                        ["234", "another", "0.456"],
                    ]
                )
            ),
        ):
            obj = FoobarEntity()
            obj.foo = 123
            obj.bar = "updated"
            obj.spam = 0.5

            class InterceptedCsvDB(CsvDB):
                def __init__(self, path):
                    CsvDB.__init__(self, path)
                    self._remained = list()

                def store(self, obj, distinct=False):
                    self._remained.append(obj)

            csvdb = InterceptedCsvDB("/foobar")
            csvdb.open()
            csvdb.create_table_from_object = MagicMock()
            csvdb.flush = MagicMock()

            assert csvdb.update(obj, eq={"foo": 123}) is True
            assert len(csvdb._remained) == 2

            assert csvdb._remained[0].foo == 123
            assert csvdb._remained[0].bar == "updated"
            assert csvdb._remained[0].spam == 0.5

            assert csvdb._remained[1].foo == 234
            assert csvdb._remained[1].bar == "another"
            assert csvdb._remained[1].spam == 0.456

    def test_get_object(self):
        """
        Getting an object from the store.
        :return:
        """
        with patch("os.listdir", MagicMock(return_value=["test_db"])), patch(
            "gzip.open", MagicMock()
        ) as gzip_mock_open, patch(
            "csv.reader",
            MagicMock(
                return_value=iter(
                    [
                        [],
                        ["foo:int", "bar:str", "spam:float"],
                        ["123", "test", "0.123"],
                        ["234", "another", "0.456"],
                    ]
                )
            ),
        ):
            csvdb = CsvDB("/foobar")
            csvdb.open()
            entities = csvdb.get(FoobarEntity)

            # test the second call to gzip.open, the first is in the open function
            assert gzip_mock_open.call_args_list[1][0][1] == "rt"

            assert list == type(entities)
            assert len(entities) == 2

            assert entities[0].foo == 123
            assert entities[0].bar == "test"
            assert entities[0].spam == 0.123

            assert entities[1].foo == 234
            assert entities[1].bar == "another"
            assert entities[1].spam == 0.456

    def test_get_obj_equals(self):
        """
        Getting an object from the store with conditions
        :return:
        """
        with patch("os.listdir", MagicMock(return_value=["test_db"])), patch(
            "gzip.open", MagicMock()
        ), patch(
            "csv.reader",
            MagicMock(
                return_value=iter(
                    [
                        [],
                        ["foo:int", "bar:str", "spam:float"],
                        ["123", "test", "0.123"],
                        ["234", "another", "0.456"],
                    ]
                )
            ),
        ):
            csvdb = CsvDB("/foobar")
            csvdb.open()

            entities = csvdb.get(FoobarEntity, eq={"foo": 123})
            assert list == type(entities)
            assert len(entities) == 1

            assert entities[0].foo == 123
            assert entities[0].bar == "test"
            assert entities[0].spam == 0.123

    def test_get_obj_more_than(self):
        """
        Getting an object from the store with conditions
        :return:
        """
        with patch("gzip.open", MagicMock()), patch(
            "os.listdir", MagicMock(return_value=["test_db"])
        ), patch(
            "csv.reader",
            MagicMock(
                return_value=iter(
                    [
                        [],
                        ["foo:int", "bar:str", "spam:float"],
                        ["123", "test", "0.123"],
                        ["234", "another", "0.456"],
                    ]
                )
            ),
        ):
            csvdb = CsvDB("/foobar")
            csvdb.open()

            entities = csvdb.get(FoobarEntity, mt={"foo": 123})
            assert list == type(entities)
            assert len(entities) == 1

            assert entities[0].foo == 234
            assert entities[0].bar == "another"
            assert entities[0].spam == 0.456

    def test_get_obj_less_than(self):
        """
        Getting an object from the store with conditions
        :return:
        """
        with patch("gzip.open", MagicMock()), patch(
            "os.listdir", MagicMock(return_value=["test_db"])
        ), patch(
            "csv.reader",
            MagicMock(
                return_value=iter(
                    [
                        [],
                        ["foo:int", "bar:str", "spam:float"],
                        ["123", "test", "0.123"],
                        ["234", "another", "0.456"],
                    ]
                )
            ),
        ):
            csvdb = CsvDB("/foobar")
            csvdb.open()

            entities = csvdb.get(FoobarEntity, lt={"foo": 234})
            assert list == type(entities)
            assert len(entities) == 1

            assert entities[0].foo == 123
            assert entities[0].bar == "test"
            assert entities[0].spam == 0.123

    def test_get_obj_matching(self):
        """
        Getting an object from the store with conditions
        :return:
        """
        with patch("gzip.open", MagicMock()), patch(
            "os.listdir", MagicMock(return_value=["test_db"])
        ), patch(
            "csv.reader",
            MagicMock(
                return_value=iter(
                    [
                        [],
                        ["foo:int", "bar:str", "spam:float"],
                        ["123", "this is test of something", "0.123"],
                        ["234", "another test of stuff", "0.456"],
                    ]
                )
            ),
        ):
            csvdb = CsvDB("/foobar")
            csvdb.open()

            entities = csvdb.get(FoobarEntity, matches={"bar": r"is\stest"})
            assert list == type(entities)
            assert len(entities) == 1

            assert entities[0].foo == 123
            assert entities[0].bar == "this is test of something"
            assert entities[0].spam == 0.123

    def test_obj_serialization(self):
        """
        Test object serialization.
        :return:
        """
        obj = FoobarEntity()
        obj.foo = 123
        obj.bar = "test entity"
        obj.spam = 0.123

        descr = OrderedDict(
            [tuple(elm.split(":")) for elm in ["foo:int", "bar:str", "spam:float"]]
        )
        assert obj._serialize(descr) == [123, "test entity", 0.123]

    def test_obj_validation(self):
        """
        Test object validation.

        :return:
        """
        with patch("os.path.exists", MagicMock(return_value=False)), patch(
            "os.listdir", MagicMock(return_value=["some_table"])
        ):
            obj = FoobarEntity()
            obj.foo = 123
            obj.bar = "test entity"
            obj.spam = 0.123

            csvdb = CsvDB("/foobar")
            csvdb._tables = {
                "some_table": OrderedDict(
                    [
                        tuple(elm.split(":"))
                        for elm in ["foo:int", "bar:str", "spam:float"]
                    ]
                )
            }
            assert csvdb._validate_object(obj) == [123, "test entity", 0.123]

    def test_criteria(self):
        """
        Test criteria selector.

        :return:
        """
        with patch("os.path.exists", MagicMock(return_value=False)), patch(
            "os.listdir", MagicMock(return_value=["some_table"])
        ):
            obj = FoobarEntity()
            obj.foo = 123
            obj.bar = "test entity"
            obj.spam = 0.123
            obj.pi = 3.14

            cmp = CsvDB("/foobar")._CsvDB__criteria

            # Single
            assert cmp(obj, eq={"foo": 123}) is True
            assert cmp(obj, lt={"foo": 124}) is True
            assert cmp(obj, mt={"foo": 122}) is True

            assert cmp(obj, eq={"foo": 0}) is False
            assert cmp(obj, lt={"foo": 123}) is False
            assert cmp(obj, mt={"foo": 123}) is False

            assert cmp(obj, matches={"bar": r"t\se.*?"}) is True
            assert cmp(obj, matches={"bar": r"\s\sentity"}) is False

            # Combined
            assert (
                cmp(obj, eq={"foo": 123, "bar": r"test entity", "spam": 0.123}) is True
            )
            assert cmp(obj, eq={"foo": 123, "bar": r"test", "spam": 0.123}) is False

            assert cmp(obj, lt={"foo": 124, "spam": 0.124}) is True
            assert cmp(obj, lt={"foo": 124, "spam": 0.123}) is False

            assert cmp(obj, mt={"foo": 122, "spam": 0.122}) is True
            assert cmp(obj, mt={"foo": 122, "spam": 0.123}) is False

            assert (
                cmp(
                    obj,
                    matches={"bar": r"test"},
                    mt={"foo": 122},
                    lt={"spam": 0.124},
                    eq={"pi": 3.14},
                )
                is True
            )

            assert (
                cmp(
                    obj,
                    matches={"bar": r"^test.*?y$"},
                    mt={"foo": 122},
                    lt={"spam": 0.124},
                    eq={"pi": 3.14},
                )
                is True
            )
            assert (
                cmp(
                    obj,
                    matches={"bar": r"^ent"},
                    mt={"foo": 122},
                    lt={"spam": 0.124},
                    eq={"pi": 3.14},
                )
                is False
            )
            assert (
                cmp(
                    obj,
                    matches={"bar": r"^test.*?y$"},
                    mt={"foo": 123},
                    lt={"spam": 0.124},
                    eq={"pi": 3.14},
                )
                is False
            )



import logging
import time

import pytest
from saltfactories.utils import random_string

import salt.cache
from salt.exceptions import SaltCacheError
from tests.pytests.functional.cache.helpers import run_common_cache_tests

pytest.importorskip("redis")
docker = pytest.importorskip("docker")

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("dockerd"),
]

pytest.importorskip("redis")


@pytest.fixture(scope="module")
def redis_container(salt_factories):
    container = salt_factories.get_container(
        random_string("redis-server-"),
        image_name="ghcr.io/saltstack/salt-ci-containers/redis:alpine",
        container_run_kwargs={"ports": {"6379/tcp": None}},
    )
    with container.started() as factory:
        yield factory


@pytest.fixture
def cache(minion_opts, redis_container):
    opts = minion_opts.copy()
    opts["cache"] = "redis"
    opts["cache.redis.host"] = "127.0.0.1"
    opts["cache.redis.port"] = redis_container.get_host_port_binding(
        6379, protocol="tcp", ipv6=False
    )
    # NOTE: If you would like to ensure that alternate prefixes are properly
    # tested, simply change these values and re-run the tests.
    opts["cache.redis.bank_prefix"] = "#BANKY_BANK"
    opts["cache.redis.bank_keys_prefix"] = "#WHO_HAS_MY_KEYS"
    opts["cache.redis.key_prefix"] = "#LPL"
    opts["cache.redis.timestamp_prefix"] = "%TICK_TOCK"
    opts["cache.redis.separator"] = "\N{SNAKE}"
    cache = salt.cache.factory(opts)
    return cache


def test_caching(subtests, cache):
    # The container seems to need some time, let's give it some
    timeout = 20
    start = time.time()
    while time.time() < start + timeout:
        try:
            cache.contains("fnord")
            break
        except SaltCacheError:
            time.sleep(1)
    else:
        pytest.fail("Failed to connect to redis container")
    run_common_cache_tests(subtests, cache)


@pytest.fixture
def redis_cluster_cache(minion_opts):
    opts = minion_opts.copy()
    opts["cache"] = "redis"
    opts["cache.redis.cluster_mode"] = True
    cache = salt.cache.factory(opts)
    yield cache


def test_redis_cluster_cache_should_import_correctly(redis_cluster_cache):
    # Doing this import here and not at the top of the module because I don't
    # want the previous test to be skipped because of this missing library.
    rediscluster_exceptions = pytest.importorskip("rediscluster.exceptions")

    with pytest.raises(rediscluster_exceptions.RedisClusterException):
        # Currently the opts aren't actually correct for a redis cluster
        # so this will fail. If, in the future, the redis_cluster_cache fixture
        # needs to point to an actual redis cluster, then this test will
        # probably become obsolete
        redis_cluster_cache.store(bank="foo", key="whatever", data="lol")



"""
Test the win_task execution module
"""

from datetime import datetime

import pytest

import salt.modules.win_task as win_task
from salt.exceptions import CommandExecutionError

pytestmark = [
    pytest.mark.skip_unless_on_windows,
    pytest.mark.destructive_test,
]


@pytest.fixture(scope="function")
def base_task():
    task_name = "SaltTest"
    result = win_task.create_task(
        task_name,
        user_name="System",
        force=True,
        action_type="Execute",
        cmd="c:\\salt\\salt-call.bat",
    )
    assert result is True
    yield task_name
    result = win_task.delete_task(task_name)
    assert result is True


def test_repeat_interval(base_task):
    result = win_task.add_trigger(
        base_task,
        trigger_type="Daily",
        trigger_enabled=True,
        repeat_duration="30 minutes",
        repeat_interval="30 minutes",
    )
    assert result is True

    result = win_task.info(base_task)
    assert result["triggers"][0]["enabled"] is True
    assert result["triggers"][0]["trigger_type"] == "Daily"
    assert result["triggers"][0]["repeat_duration"] == "30 minutes"
    assert result["triggers"][0]["repeat_interval"] == "30 minutes"


def test_repeat_interval_and_indefinitely(base_task):
    result = win_task.add_trigger(
        base_task,
        trigger_type="Daily",
        trigger_enabled=True,
        repeat_duration="Indefinitely",
        repeat_interval="30 minutes",
    )
    assert result is True

    result = win_task.info(base_task)
    assert result["triggers"][0]["enabled"] is True
    assert result["triggers"][0]["trigger_type"] == "Daily"
    assert result["triggers"][0]["repeat_duration"] == "Indefinitely"
    assert result["triggers"][0]["repeat_interval"] == "30 minutes"


def test_edit_task_delete_after(base_task):
    result = win_task.add_trigger(
        base_task,
        trigger_type="Daily",
        trigger_enabled=True,
        end_date=datetime.today().strftime("%Y-%m-%d"),
        end_time="23:59:59",
    )
    assert result is True

    result = win_task.edit_task(base_task, delete_after="30 days")
    assert result is True

    result = win_task.info(base_task)
    assert result["settings"]["delete_after"] == "30 days"

    result = win_task.edit_task(base_task, delete_after=False)
    assert result is True

    result = win_task.info(base_task)
    assert result["settings"]["delete_after"] is False


def test_execution_time_limit(base_task):
    result = win_task.add_trigger(
        base_task,
        trigger_type="Daily",
        trigger_enabled=True,
        end_date=datetime.today().strftime("%Y-%m-%d"),
        end_time="23:59:59",
    )
    assert result is True

    result = win_task.edit_task(base_task, execution_time_limit="1 hour")
    assert result is True

    result = win_task.info(base_task)
    assert result["settings"]["execution_time_limit"] == "1 hour"

    result = win_task.edit_task(base_task, execution_time_limit=False)
    assert result is True

    result = win_task.info(base_task)
    assert result["settings"]["execution_time_limit"] is False


@pytest.mark.parametrize(
    "exitcode, expect",
    [
        (0, "The operation completed successfully"),
        (3221225786, "The application terminated as a result of CTRL+C"),
        (4289449455, "Unknown Task Result: 0xffabcdef"),
    ],
)
def test_run_result_code(exitcode, expect):
    task_name = "SaltTest"
    try:
        result = win_task.create_task(
            task_name,
            user_name="System",
            force=True,
            action_type="Execute",
            cmd="cmd.exe",
            arguments=f"/c exit {exitcode}",
        )
        assert result is True

        result = win_task.info(task_name)
        assert result["last_run_result"] == "Task has not yet run"

        result = win_task.run_wait(task_name)
        assert result is True

        result = win_task.info(task_name)
        assert result["last_run_result"] == expect
    finally:
        result = win_task.delete_task(task_name)
        assert result is True


def test_create_task_from_xml():
    task_name = "SaltTest"
    task_xml = '<Task xmlns="http://schemas.microsoft.com/windows/2004/02/mit/task"><Actions><Exec><Command>cmd.exe</Command><Arguments>/c exit</Arguments></Exec></Actions></Task>'
    try:
        result = win_task.create_task_from_xml(
            task_name, user_name="System", xml_text=task_xml
        )
        assert result is True

        result = win_task.info(task_name)
        assert result["actions"][0]["action_type"] == "Execute"
        assert result["actions"][0]["cmd"] == "cmd.exe"
        assert result["actions"][0]["arguments"] == "/c exit"

    finally:
        result = win_task.delete_task(task_name)
        assert result is True


def test_create_task_from_xml_error():
    task_name = "SaltTest"
    try:
        with pytest.raises(CommandExecutionError) as excinfo:
            result = win_task.create_task_from_xml(
                task_name, user_name="System", xml_text="test"
            )
            assert result is False
        assert "The task XML is malformed" in str(excinfo.value)
    finally:
        result = win_task.delete_task(task_name)
        assert result is not True



"""
Control the Salt command interface
==================================

This state is intended for use from the Salt Master. It provides access to
sending commands down to minions as well as access to executing master-side
modules. These state functions wrap Salt's :ref:`Python API <python-api>`.

    .. versionadded:: 2016.11.0

    Support for masterless minions was added to the ``salt.state`` function,
    so they can run orchestration sls files. This is particularly useful when
    the rendering of a state is dependent on the execution of another state.
    Orchestration will render and execute each orchestration block
    independently, while honoring requisites to ensure the states are applied
    in the correct order.

.. seealso:: More Orchestrate documentation

    * :ref:`Full Orchestrate Tutorial <orchestrate-runner>`
    * :py:func:`The Orchestrate runner <salt.runners.state.orchestrate>`
"""

import fnmatch
import logging
import sys
import threading
import time

import salt.exceptions
import salt.output
import salt.syspaths
import salt.utils.data
import salt.utils.event
import salt.utils.versions

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "salt"


def __virtual__():
    """
    Named salt
    """
    return __virtualname__


def _fire_args(tag_data):
    try:
        salt.utils.event.fire_args(__opts__, __orchestration_jid__, tag_data, "run")
    except NameError:
        log.debug("Unable to fire args event due to missing __orchestration_jid__")


def _parallel_map(func, inputs):
    """
    Applies a function to each element of a list, returning the resulting list.

    A separate thread is created for each element in the input list and the
    passed function is called for each of the elements. When all threads have
    finished execution a list with the results corresponding to the inputs is
    returned.

    If one of the threads fails (because the function throws an exception),
    that exception is reraised. If more than one thread fails, the exception
    from the first thread (according to the index of the input element) is
    reraised.

    func:
        function that is applied on each input element.
    inputs:
        list of elements that shall be processed. The length of this list also
        defines the number of threads created.
    """
    outputs = len(inputs) * [None]
    errors = len(inputs) * [None]

    def create_thread(index):
        def run_thread():
            try:
                outputs[index] = func(inputs[index])
            except:  # pylint: disable=bare-except
                errors[index] = sys.exc_info()

        thread = threading.Thread(target=run_thread)
        thread.start()
        return thread

    threads = list(map(create_thread, range(len(inputs))))
    for thread in threads:
        thread.join()
    for error in errors:
        if error is not None:
            exc_type, exc_value, exc_traceback = error
            raise exc_value.with_traceback(exc_traceback)
    return outputs


def state(
    name,
    tgt,
    ssh=False,
    tgt_type="glob",
    ret="",
    ret_config=None,
    ret_kwargs=None,
    highstate=None,
    sls=None,
    top=None,
    saltenv=None,
    test=None,
    pillar=None,
    pillarenv=None,
    expect_minions=True,
    exclude=None,
    fail_minions=None,
    allow_fail=0,
    concurrent=False,
    timeout=None,
    batch=None,
    queue=False,
    subset=None,
    orchestration_jid=None,
    failhard=None,
    **kwargs,
):
    """
    Invoke a state run on a given target

    name
        An arbitrary name used to track the state execution

    tgt
        The target specification for the state run.

        .. versionadded:: 2016.11.0

        Masterless support: When running on a masterless minion, the ``tgt``
        is ignored and will always be the local minion.

    tgt_type
        The target type to resolve, defaults to ``glob``

    ret
        Optionally set a single or a list of returners to use

    ret_config
        Use an alternative returner configuration

    ret_kwargs
        Override individual returner configuration items

    highstate
        Defaults to None, if set to True the target systems will ignore any
        sls references specified in the sls option and call state.highstate
        on the targeted minions

    top
        Should be the name of a top file. If set state.top is called with this
        top file instead of state.sls.

    sls
        A group of sls files to execute. This can be defined as a single string
        containing a single sls file, or a list of sls files

    test
        Pass ``test=true`` or ``test=false`` through to the state function. This
        can be used to override a test mode set in the minion's config file. If
        left as the default of None and the 'test' mode is supplied on the
        command line, that value is passed instead.

    pillar
        Pass the ``pillar`` kwarg through to the state function

    pillarenv
        The pillar environment to grab pillars from

        .. versionadded:: 2017.7.0

    saltenv
        The default salt environment to pull sls files from

    ssh
        Set to `True` to use the ssh client instead of the standard salt client

    roster
        In the event of using salt-ssh, a roster system can be set

    expect_minions
        An optional boolean for failing if some minions do not respond

    fail_minions
        An optional list of targeted minions where failure is an option

    allow_fail
        Pass in the number of minions to allow for failure before setting
        the result of the execution to False

    exclude
        Pass exclude kwarg to state

    concurrent
        Allow multiple state runs to occur at once.

        WARNING: This flag is potentially dangerous. It is designed
        for use when multiple state runs can safely be run at the same
        Do not use this flag for performance optimization.

    queue
        Pass ``queue=true`` through to the state function

    batch
        Execute the command :ref:`in batches <targeting-batch>`. E.g.: ``10%``.

        .. versionadded:: 2016.3.0

    subset
        Number of minions from the targeted set to randomly use

        .. versionadded:: 2017.7.0

    failhard
        pass failhard down to the executing state

        .. versionadded:: 2019.2.2

    Examples:

    Run a list of sls files via :py:func:`state.sls <salt.state.sls>` on target
    minions:

    .. code-block:: yaml

        webservers:
          salt.state:
            - tgt: 'web*'
            - sls:
              - apache
              - django
              - core
            - saltenv: prod

    Run sls file via :py:func:`state.sls <salt.state.sls>` on target
    minions with exclude:

    .. code-block:: yaml

        docker:
          salt.state:
            - tgt: 'docker*'
            - sls: docker
            - exclude: docker.swarm
            - saltenv: prod

    Run a full :py:func:`state.highstate <salt.state.highstate>` on target
    mininons.

    .. code-block:: yaml

        databases:
          salt.state:
            - tgt: role:database
            - tgt_type: grain
            - highstate: True
    """
    cmd_kw = {"arg": [], "kwarg": {}, "ret": ret, "timeout": timeout}

    if ret_config:
        cmd_kw["ret_config"] = ret_config

    if ret_kwargs:
        cmd_kw["ret_kwargs"] = ret_kwargs

    state_ret = {"name": name, "changes": {}, "comment": "", "result": True}

    try:
        allow_fail = int(allow_fail)
    except ValueError:
        state_ret["result"] = False
        state_ret["comment"] = "Passed invalid value for 'allow_fail', must be an int"
        return state_ret

    cmd_kw["tgt_type"] = tgt_type
    cmd_kw["ssh"] = ssh
    if "roster" in kwargs:
        cmd_kw["roster"] = kwargs["roster"]
    cmd_kw["expect_minions"] = expect_minions
    if highstate:
        fun = "state.highstate"
    elif top:
        fun = "state.top"
        cmd_kw["arg"].append(top)
    elif sls:
        fun = "state.sls"
        if isinstance(sls, list):
            sls = ",".join(sls)
        cmd_kw["arg"].append(sls)
    else:
        state_ret["comment"] = "No highstate or sls specified, no execution made"
        state_ret["result"] = False
        return state_ret

    if test is not None or __opts__.get("test"):
        cmd_kw["kwarg"]["test"] = test if test is not None else __opts__.get("test")

    if pillar:
        cmd_kw["kwarg"]["pillar"] = pillar

    if pillarenv is not None:
        cmd_kw["kwarg"]["pillarenv"] = pillarenv

    if saltenv is not None:
        cmd_kw["kwarg"]["saltenv"] = saltenv

    if exclude is not None:
        cmd_kw["kwarg"]["exclude"] = exclude

    cmd_kw["kwarg"]["queue"] = queue

    if isinstance(concurrent, bool):
        cmd_kw["kwarg"]["concurrent"] = concurrent
    else:
        state_ret["comment"] = "Must pass in boolean for value of 'concurrent'"
        state_ret["result"] = False
        return state_ret

    if batch is not None:
        cmd_kw["batch"] = str(batch)

    if subset is not None:
        cmd_kw["subset"] = subset

    if failhard is True or __opts__.get("failhard"):
        cmd_kw["failhard"] = True

    masterless = __opts__["__role"] == "minion" and __opts__["file_client"] == "local"
    if not masterless:
        _fire_args({"type": "state", "tgt": tgt, "name": name, "args": cmd_kw})
        cmd_ret = __salt__["saltutil.cmd"](tgt, fun, **cmd_kw)
    else:
        if top:
            cmd_kw["topfn"] = "".join(cmd_kw.pop("arg"))
        elif sls:
            cmd_kw["mods"] = "".join(cmd_kw.pop("arg"))
        cmd_kw.update(cmd_kw.pop("kwarg"))
        tmp_ret = __salt__[fun](**cmd_kw)
        cmd_ret = {
            __opts__["id"]: {
                "ret": tmp_ret,
                "out": (
                    tmp_ret.get("out", "highstate")
                    if isinstance(tmp_ret, dict)
                    else "highstate"
                ),
            }
        }

    try:
        state_ret["__jid__"] = cmd_ret[next(iter(cmd_ret))]["jid"]
    except (StopIteration, KeyError):
        pass

    changes = {}
    fail = set()
    no_change = set()

    if fail_minions is None:
        fail_minions = ()
    elif isinstance(fail_minions, str):
        fail_minions = [minion.strip() for minion in fail_minions.split(",")]
    elif not isinstance(fail_minions, list):
        state_ret.setdefault("warnings", []).append(
            "'fail_minions' needs to be a list or a comma separated string. Ignored."
        )
        fail_minions = ()

    if not cmd_ret and expect_minions:
        state_ret["result"] = False
        state_ret["comment"] = "No minions returned"
        return state_ret

    for minion, mdata in cmd_ret.items():
        if mdata.get("out", "") != "highstate":
            log.warning("Output from salt state not highstate")

        m_ret = False

        if "return" in mdata and "ret" not in mdata:
            mdata["ret"] = mdata.pop("return")

        m_state = True
        if mdata.get("failed", False):
            m_state = False
        else:
            try:
                m_ret = mdata["ret"]
            except KeyError:
                m_state = False
            if m_state:
                m_state = __utils__["state.check_result"](m_ret, recurse=True)

        if not m_state:
            if minion not in fail_minions:
                fail.add(minion)
            changes[minion] = m_ret
            continue
        try:
            for state_item in m_ret.values():
                if isinstance(state_item, dict):
                    if "changes" in state_item and state_item["changes"]:
                        changes[minion] = m_ret
                        break
            else:
                no_change.add(minion)
        except AttributeError:
            log.error("m_ret did not have changes %s %s", type(m_ret), m_ret)
            no_change.add(minion)

    if changes:
        state_ret["changes"] = {"out": "highstate", "ret": changes}
    if len(fail) > allow_fail:
        state_ret["result"] = False
        state_ret["comment"] = "Run failed on minions: {}".format(", ".join(fail))
    else:
        state_ret["comment"] = "States ran successfully."
        if changes:
            state_ret["comment"] += " Updating {}.".format(", ".join(changes))
        if no_change:
            state_ret["comment"] += " No changes made to {}.".format(
                ", ".join(no_change)
            )
    if test or __opts__.get("test"):
        if state_ret["changes"] and state_ret["result"] is True:
            # Test mode with changes is the only case where result should ever be none
            state_ret["result"] = None
    return state_ret


def function(
    name,
    tgt,
    ssh=False,
    tgt_type="glob",
    ret="",
    ret_config=None,
    ret_kwargs=None,
    expect_minions=False,
    fail_minions=None,
    fail_function=None,
    arg=None,
    kwarg=None,
    timeout=None,
    batch=None,
    subset=None,
    failhard=None,
    **kwargs,
):  # pylint: disable=unused-argument
    """
    Execute a single module function on a remote minion via salt or salt-ssh

    name
        The name of the function to run, aka cmd.run or pkg.install

    tgt
        The target specification, aka '*' for all minions

    tgt_type
        The target type, defaults to ``glob``

    arg
        The list of arguments to pass into the function

    kwarg
        The dict (not a list) of keyword arguments to pass into the function

    ret
        Optionally set a single or a list of returners to use

    ret_config
        Use an alternative returner configuration

    ret_kwargs
        Override individual returner configuration items

    expect_minions
        An optional boolean for failing if some minions do not respond

    fail_minions
        An optional list of targeted minions where failure is an option

    fail_function
        An optional string that points to a salt module that returns True or False
        based on the returned data dict for individual minions

    ssh
        Set to `True` to use the ssh client instead of the standard salt client

    roster
        In the event of using salt-ssh, a roster system can be set

        .. versionadded:: 3005

    batch
        Execute the command :ref:`in batches <targeting-batch>`. E.g.: ``10%``.

    subset
        Number of minions from the targeted set to randomly use

        .. versionadded:: 2017.7.0

    failhard
        pass failhard down to the executing state

        .. versionadded:: 2019.2.2

    """
    func_ret = {"name": name, "changes": {}, "comment": "", "result": True}
    if kwarg is None:
        kwarg = {}
    if isinstance(arg, str):
        func_ret["warnings"] = ["Please specify 'arg' as a list of arguments."]
        arg = arg.split()

    cmd_kw = {"arg": arg or [], "kwarg": kwarg, "ret": ret, "timeout": timeout}

    if batch is not None:
        cmd_kw["batch"] = str(batch)
    if subset is not None:
        cmd_kw["subset"] = subset

    cmd_kw["tgt_type"] = tgt_type
    cmd_kw["ssh"] = ssh
    if "roster" in kwargs:
        cmd_kw["roster"] = kwargs["roster"]
    cmd_kw["expect_minions"] = expect_minions
    cmd_kw["_cmd_meta"] = True

    if failhard is True or __opts__.get("failhard"):
        cmd_kw["failhard"] = True

    if ret_config:
        cmd_kw["ret_config"] = ret_config

    if ret_kwargs:
        cmd_kw["ret_kwargs"] = ret_kwargs

    fun = name
    if __opts__["test"] is True:
        func_ret["comment"] = "Function {} would be executed on target {}".format(
            fun, tgt
        )
        func_ret["result"] = None
        return func_ret
    try:
        _fire_args({"type": "function", "tgt": tgt, "name": name, "args": cmd_kw})
        cmd_ret = __salt__["saltutil.cmd"](tgt, fun, **cmd_kw)
    except Exception as exc:  # pylint: disable=broad-except
        func_ret["result"] = False
        func_ret["comment"] = str(exc)
        return func_ret

    try:
        func_ret["__jid__"] = cmd_ret[next(iter(cmd_ret))]["jid"]
    except (StopIteration, KeyError):
        pass

    changes = {}
    fail = set()

    if fail_minions is None:
        fail_minions = ()
    elif isinstance(fail_minions, str):
        fail_minions = [minion.strip() for minion in fail_minions.split(",")]
    elif not isinstance(fail_minions, list):
        func_ret.setdefault("warnings", []).append(
            "'fail_minions' needs to be a list or a comma separated string. Ignored."
        )
        fail_minions = ()
    for minion, mdata in cmd_ret.items():
        m_ret = False
        if mdata.get("retcode"):
            func_ret["result"] = False
            fail.add(minion)
        if mdata.get("failed", False):
            m_func = False
        else:
            if "return" in mdata and "ret" not in mdata:
                mdata["ret"] = mdata.pop("return")
            m_ret = mdata["ret"]
            m_func = (not fail_function and True) or __salt__[fail_function](m_ret)

            if m_ret is False:
                m_func = False

        if not m_func:
            if minion not in fail_minions:
                fail.add(minion)
        changes[minion] = m_ret
    if not cmd_ret:
        func_ret["result"] = False
        func_ret["comment"] = "No minions responded"
    else:
        if changes:
            func_ret["changes"] = {"ret": changes}
        if fail:
            func_ret["result"] = False
            func_ret["comment"] = "Running function {} failed on minions: {}".format(
                name, ", ".join(fail)
            )
        else:
            func_ret["comment"] = "Function ran successfully."
        if changes:
            func_ret["comment"] += " Function {} ran on {}.".format(
                name, ", ".join(changes)
            )
    return func_ret


def wait_for_event(name, id_list, event_id="id", timeout=300, node="master"):
    """
    Watch Salt's event bus and block until a condition is met

    .. versionadded:: 2014.7.0

    name
        An event tag to watch for; supports Reactor-style globbing.
    id_list
        A list of event identifiers to watch for -- usually the minion ID. Each
        time an event tag is matched the event data is inspected for
        ``event_id``, if found it is removed from ``id_list``. When ``id_list``
        is empty this function returns success.
    event_id : id
        The name of a key in the event data. Default is ``id`` for the minion
        ID, another common value is ``name`` for use with orchestrating
        salt-cloud events.
    timeout : 300
        The maximum time in seconds to wait before failing.

    The following example blocks until all the listed minions complete a
    restart and reconnect to the Salt master:

    .. code-block:: yaml

        reboot_all_minions:
          salt.function:
            - name: system.reboot
            - tgt: '*'

        wait_for_reboots:
          salt.wait_for_event:
            - name: salt/minion/*/start
            - id_list:
              - jerry
              - stuart
              - dave
              - phil
              - kevin
              - mike
            - require:
              - salt: reboot_all_minions
    """
    ret = {"name": name, "changes": {}, "comment": "", "result": False}

    if __opts__.get("test"):
        ret["comment"] = f"Orchestration would wait for event '{name}'"
        ret["result"] = None
        return ret

    with salt.utils.event.get_event(
        node, __opts__["sock_dir"], opts=__opts__, listen=True
    ) as sevent:

        del_counter = 0
        starttime = time.time()
        timelimit = starttime + timeout
        while True:
            event = sevent.get_event(full=True)
            is_timedout = time.time() > timelimit

            if event is None and not is_timedout:
                log.trace("wait_for_event: No event data; waiting.")
                continue
            elif event is None and is_timedout:
                ret["comment"] = "Timeout value reached."
                return ret

            if fnmatch.fnmatch(event["tag"], name):
                val = event["data"].get(event_id)
                if val is None and "data" in event["data"]:
                    val = event["data"]["data"].get(event_id)

                if val is not None:
                    if isinstance(val, list):

                        val_list = [id for id in id_list if id in val]

                        if not val_list:
                            log.trace(
                                "wait_for_event: Event identifier '%s' not in "
                                "id_list; skipping",
                                event_id,
                            )
                        elif val_list:
                            minions_seen = ret["changes"].setdefault("minions_seen", [])
                            for found_val in val_list:
                                id_list.remove(found_val)
                                del_counter += 1
                                minions_seen.append(found_val)
                                log.debug(
                                    "wait_for_event: Event identifier '%s' removed "
                                    "from id_list; %s items remaining.",
                                    found_val,
                                    len(id_list),
                                )

                    else:
                        try:
                            val_idx = id_list.index(val)
                        except ValueError:
                            log.trace(
                                "wait_for_event: Event identifier '%s' not in "
                                "id_list; skipping.",
                                event_id,
                            )
                        else:
                            del id_list[val_idx]
                            del_counter += 1
                            minions_seen = ret["changes"].setdefault("minions_seen", [])
                            minions_seen.append(val)

                            log.debug(
                                "wait_for_event: Event identifier '%s' removed "
                                "from id_list; %s items remaining.",
                                val,
                                len(id_list),
                            )
                else:
                    log.trace(
                        "wait_for_event: Event identifier '%s' not in event "
                        "'%s'; skipping.",
                        event_id,
                        event["tag"],
                    )
            else:
                log.debug("wait_for_event: Skipping unmatched event '%s'", event["tag"])

            if len(id_list) == 0:
                ret["result"] = True
                ret["comment"] = "All events seen in {} seconds.".format(
                    time.time() - starttime
                )
                return ret

            if is_timedout:
                ret["comment"] = "Timeout value reached."
                return ret


def runner(name, **kwargs):
    """
    Execute a runner module on the master

    .. versionadded:: 2014.7.0

    name
        The name of the function to run
    kwargs
        Any keyword arguments to pass to the runner function

    .. code-block:: yaml

         run-manage-up:
          salt.runner:
            - name: manage.up
    """
    try:
        jid = __orchestration_jid__
    except NameError:
        log.debug("Unable to fire args event due to missing __orchestration_jid__")
        jid = None

    try:
        kwargs["__pub_user"] = __user__
        log.debug(
            "added __pub_user to kwargs using dunder user '%s', kwargs '%s'",
            __user__,
            kwargs,
        )
    except NameError:
        log.warning("unable to find user for fire args event due to missing __user__")

    if __opts__.get("test", False):
        ret = {
            "name": name,
            "result": None,
            "changes": {},
            "comment": f"Runner function '{name}' would be executed.",
        }
        return ret

    out = __salt__["saltutil.runner"](
        name, __orchestration_jid__=jid, __env__=__env__, full_return=True, **kwargs
    )

    runner_return = out.get("return")
    if isinstance(runner_return, dict) and "Error" in runner_return:
        out["success"] = False

    success = out.get("success", True)
    ret = {"name": name, "changes": {"return": runner_return}, "result": success}
    ret["comment"] = "Runner function '{}' {}.".format(
        name,
        "executed" if success else "failed",
    )

    if __opts__["features"].get("enable_deprecated_orchestration_flag", False):
        ret["__orchestration__"] = True
        salt.utils.versions.warn_until(
            3008,
            "The __orchestration__ return flag will be removed in {version}. "
            "For more information see https://github.com/saltstack/salt/pull/59917.",
        )

    if "jid" in out:
        ret["__jid__"] = out["jid"]

    return ret


def parallel_runners(name, runners, **kwargs):  # pylint: disable=unused-argument
    """
    Executes multiple runner modules on the master in parallel.

    .. versionadded:: 2018.3.0

    A separate thread is spawned for each runner. This state is intended to be
    used with the orchestrate runner in place of the ``saltmod.runner`` state
    when different tasks should be run in parallel. In general, Salt states are
    not safe when used concurrently, so ensure that they are used in a safe way
    (e.g. by only targeting separate minions in parallel tasks).

    name:
        name identifying this state. The name is provided as part of the
        output, but not used for anything else.

    runners:
        list of runners that should be run in parallel. Each element of the
        list has to be a dictionary. This dictionary's name entry stores the
        name of the runner function that shall be invoked. The optional kwarg
        entry stores a dictionary of named arguments that are passed to the
        runner function.

    .. code-block:: yaml

        parallel-state:
           salt.parallel_runners:
             - runners:
                 my_runner_1:
                   - name: state.orchestrate
                   - kwarg:
                       mods: orchestrate_state_1
                 my_runner_2:
                   - name: state.orchestrate
                   - kwarg:
                       mods: orchestrate_state_2
    """
    # For the sake of consistency, we treat a single string in the same way as
    # a key without a value. This allows something like
    #     salt.parallel_runners:
    #       - runners:
    #           state.orchestrate
    # Obviously, this will only work if the specified runner does not need any
    # arguments.
    if isinstance(runners, str):
        runners = {runners: [{name: runners}]}
    # If the runners argument is not a string, it must be a dict. Everything
    # else is considered an error.
    if not isinstance(runners, dict):
        return {
            "name": name,
            "result": False,
            "changes": {},
            "comment": "The runners parameter must be a string or dict.",
        }
    # The configuration for each runner is given as a list of key-value pairs.
    # This is not very useful for what we want to do, but it is the typical
    # style used in Salt. For further processing, we convert each of these
    # lists to a dict. This also makes it easier to check whether a name has
    # been specified explicitly.
    for runner_id, runner_config in runners.items():
        if runner_config is None:
            runner_config = {}
        else:
            runner_config = salt.utils.data.repack_dictlist(runner_config)
        if "name" not in runner_config:
            runner_config["name"] = runner_id
        runners[runner_id] = runner_config

    try:
        jid = __orchestration_jid__
    except NameError:
        log.debug("Unable to fire args event due to missing __orchestration_jid__")
        jid = None

    def call_runner(runner_config):
        return __salt__["saltutil.runner"](
            runner_config["name"],
            __orchestration_jid__=jid,
            __env__=__env__,
            full_return=True,
            **(runner_config.get("kwarg", {})),
        )

    try:
        outputs = _parallel_map(call_runner, list(runners.values()))
    except salt.exceptions.SaltException as exc:
        return {
            "name": name,
            "result": False,
            "success": False,
            "changes": {},
            "comment": f"One of the runners raised an exception: {exc}",
        }
    # We bundle the results of the runners with the IDs of the runners so that
    # we can easily identify which output belongs to which runner. At the same
    # time we exctract the actual return value of the runner (saltutil.runner
    # adds some extra information that is not interesting to us).
    outputs = {
        runner_id: out["return"] for runner_id, out in zip(runners.keys(), outputs)
    }

    # If each of the runners returned its output in the format compatible with
    # the 'highstate' outputter, we can leverage this fact when merging the
    # outputs.
    highstate_output = all(
        [
            out.get("outputter", "") == "highstate" and "data" in out
            for out in outputs.values()
        ]
    )

    # The following helper function is used to extract changes from highstate
    # output.

    def extract_changes(obj):
        if not isinstance(obj, dict):
            return {}
        elif "changes" in obj:
            if (
                isinstance(obj["changes"], dict)
                and obj["changes"].get("out", "") == "highstate"
                and "ret" in obj["changes"]
            ):
                return obj["changes"]["ret"]
            else:
                return obj["changes"]
        else:
            found_changes = {}
            for key, value in obj.items():
                change = extract_changes(value)
                if change:
                    found_changes[key] = change
            return found_changes

    if highstate_output:
        failed_runners = [
            runner_id
            for runner_id, out in outputs.items()
            if out["data"].get("retcode", 0) != 0
        ]
        all_successful = not failed_runners
        if all_successful:
            comment = "All runner functions executed successfully."
        else:
            runner_comments = [
                "Runner {} failed with return value:\n{}".format(
                    runner_id,
                    salt.output.out_format(
                        outputs[runner_id], "nested", __opts__, nested_indent=2
                    ),
                )
                for runner_id in failed_runners
            ]
            comment = "\n".join(runner_comments)
        changes = {}
        for runner_id, out in outputs.items():
            runner_changes = extract_changes(out["data"])
            if runner_changes:
                changes[runner_id] = runner_changes
    else:
        failed_runners = [
            runner_id
            for runner_id, out in outputs.items()
            if out.get("exit_code", 0) != 0
        ]
        all_successful = not failed_runners
        if all_successful:
            comment = "All runner functions executed successfully."
        else:
            if len(failed_runners) == 1:
                comment = f"Runner {failed_runners[0]} failed."
            else:
                comment = "Runners {} failed.".format(", ".join(failed_runners))
        changes = {"ret": {runner_id: out for runner_id, out in outputs.items()}}
    ret = {
        "name": name,
        "result": all_successful,
        "changes": changes,
        "comment": comment,
    }

    # The 'runner' function includes out['jid'] as '__jid__' in the returned
    # dict, but we cannot do this here because we have more than one JID if
    # we have more than one runner.

    return ret


def wheel(name, **kwargs):
    """
    Execute a wheel module on the master

    .. versionadded:: 2014.7.0

    name
        The name of the function to run
    kwargs
        Any keyword arguments to pass to the wheel function

    .. code-block:: yaml

        accept_minion_key:
          salt.wheel:
            - name: key.accept
            - match: frank
    """
    ret = {"name": name, "result": False, "changes": {}, "comment": ""}
    try:
        jid = __orchestration_jid__
    except NameError:
        log.debug("Unable to fire args event due to missing __orchestration_jid__")
        jid = None

    if __opts__.get("test", False):
        ret["result"] = None
        ret["changes"] = {}
        ret["comment"] = f"Wheel function '{name}' would be executed."
        return ret

    out = __salt__["saltutil.wheel"](
        name, __orchestration_jid__=jid, __env__=__env__, **kwargs
    )

    wheel_return = out.get("return")
    if isinstance(wheel_return, dict) and "Error" in wheel_return:
        out["success"] = False

    success = out.get("success", True)
    ret = {"name": name, "changes": {"return": wheel_return}, "result": success}
    ret["comment"] = "Wheel function '{}' {}.".format(
        name,
        "executed" if success else "failed",
    )

    if __opts__["features"].get("enable_deprecated_orchestration_flag", False):
        ret["__orchestration__"] = True
        salt.utils.versions.warn_until(
            "Argon",
            "The __orchestration__ return flag will be removed in Salt Argon. "
            "For more information see https://github.com/saltstack/salt/pull/59917.",
        )

    if "jid" in out:
        ret["__jid__"] = out["jid"]

    return ret



import pytest

pytestmark = [
    pytest.mark.skip_on_windows,
]


@pytest.mark.parametrize("output_fmt", ["yaml", "json"])
def test_salt_output(salt_cli, salt_minion, salt_master, output_fmt):
    """
    Test --output
    """
    assert salt_master.is_running()

    ret = salt_cli.run(
        f"--output={output_fmt}", "test.fib", "7", minion_tgt=salt_minion.id
    )
    if output_fmt == "json":
        assert 13 in ret.data
    else:
        ret.stdout.matcher.fnmatch_lines(["*- 13*"])



import os

import pytest


@pytest.fixture(scope="module")
def install():
    pytest.helpers.clean_env()
    # Create old config
    pytest.helpers.old_install()
    # Create a custom config
    pytest.helpers.custom_config()
    args = ["/S", "/custom-config=custom_conf", "/move-config"]
    pytest.helpers.install_salt(args)
    yield args
    pytest.helpers.clean_env()


def test_binaries_present_old_location(install):
    # This will show the contents of the directory on failure
    dir_contents = os.listdir(rf"{pytest.OLD_DIR}\bin")
    # Apparently we don't move the binaries even if they pass install-dir
    # TODO: Decide if this is expected behavior
    assert os.path.exists(rf"{pytest.OLD_DIR}\bin\ssm.exe")
    assert os.path.exists(rf"{pytest.OLD_DIR}\bin\python.exe")


def test_config_present_new_location(install):
    assert os.path.exists(rf"{pytest.DATA_DIR}\conf\minion")


def test_config_correct(install):
    # The config file should be the custom config in the new location, unchanged
    with open(rf"{pytest.SCRIPT_DIR}\custom_conf") as f:
        expected = f.readlines()

    with open(rf"{pytest.DATA_DIR}\conf\minion") as f:
        result = f.readlines()

    assert result == expected



# isort:skip_file

import os
import sys

sys.modules["pkg_resources"] = None

import salt.defaults.exitcodes
import salt.utils.parsers
import salt.utils.stringutils
from salt.exceptions import (
    AuthenticationError,
    AuthorizationError,
    EauthAuthenticationError,
    LoaderError,
    SaltClientError,
    SaltInvocationError,
    SaltSystemExit,
)
from salt.utils.args import yamlify_arg


class SaltCMD(salt.utils.parsers.SaltCMDOptionParser):
    """
    The execution of a salt command happens here
    """

    def run(self):
        """
        Execute the salt command line
        """
        import salt.client

        self.parse_args()

        try:
            # We don't need to bail on config file permission errors
            # if the CLI process is run with the -a flag
            skip_perm_errors = self.options.eauth != ""

            self.local_client = salt.client.get_local_client(
                self.get_config_file_path(),
                skip_perm_errors=skip_perm_errors,
                auto_reconnect=True,
            )
        except SaltClientError as exc:
            self.exit(2, f"{exc}\n")
            return

        if self.options.batch or self.options.static:
            # _run_batch() will handle all output and
            # exit with the appropriate error condition
            # Execution will not continue past this point
            # in batch mode.
            self._run_batch()
            return

        if self.options.preview_target:
            minion_list = self._preview_target()
            self._output_ret(minion_list, self.config.get("output", "nested"))
            return

        if self.options.timeout <= 0:
            self.options.timeout = self.local_client.opts["timeout"]

        kwargs = {
            "tgt": self.config["tgt"],
            "fun": self.config["fun"],
            "arg": self.config["arg"],
            "timeout": self.options.timeout,
            "show_timeout": self.options.show_timeout,
            "show_jid": self.options.show_jid,
        }

        if "token" in self.config:
            import salt.utils.files

            try:
                with salt.utils.files.fopen(
                    os.path.join(self.config["cachedir"], ".root_key"), "r"
                ) as fp_:
                    kwargs["key"] = fp_.readline()
            except OSError:
                kwargs["token"] = self.config["token"]

        kwargs["delimiter"] = self.options.delimiter

        if self.selected_target_option:
            kwargs["tgt_type"] = self.selected_target_option
        else:
            kwargs["tgt_type"] = "glob"

        # If batch_safe_limit is set, check minions matching target and
        # potentially switch to batch execution
        if self.options.batch_safe_limit > 1:
            if len(self._preview_target()) >= self.options.batch_safe_limit:
                salt.utils.stringutils.print_cli(
                    "\nNOTICE: Too many minions targeted, switching to batch execution."
                )
                self.options.batch = self.options.batch_safe_size
                try:
                    self._run_batch()
                finally:
                    self.local_client.destroy()
                return

        if getattr(self.options, "return"):
            kwargs["ret"] = getattr(self.options, "return")

        if getattr(self.options, "return_config"):
            kwargs["ret_config"] = getattr(self.options, "return_config")

        if getattr(self.options, "return_kwargs"):
            kwargs["ret_kwargs"] = yamlify_arg(getattr(self.options, "return_kwargs"))

        if getattr(self.options, "module_executors"):
            kwargs["module_executors"] = yamlify_arg(
                getattr(self.options, "module_executors")
            )

        if getattr(self.options, "executor_opts"):
            kwargs["executor_opts"] = yamlify_arg(
                getattr(self.options, "executor_opts")
            )

        if getattr(self.options, "metadata"):
            kwargs["metadata"] = yamlify_arg(getattr(self.options, "metadata"))

        # If using eauth and a token hasn't already been loaded into
        # kwargs, prompt the user to enter auth credentials
        if "token" not in kwargs and "key" not in kwargs and self.options.eauth:
            # This is expensive. Don't do it unless we need to.
            import salt.auth

            resolver = salt.auth.Resolver(self.config)
            res = resolver.cli(self.options.eauth)
            if self.options.mktoken and res:
                tok = resolver.token_cli(self.options.eauth, res)
                if tok:
                    kwargs["token"] = tok.get("token", "")
            if not res:
                sys.stderr.write("ERROR: Authentication failed\n")
                sys.exit(2)
            kwargs.update(res)
            kwargs["eauth"] = self.options.eauth

        if self.config["async"]:
            jid = self.local_client.cmd_async(**kwargs)
            salt.utils.stringutils.print_cli(f"Executed command with job ID: {jid}")
            return

        # local will be None when there was an error
        if not self.local_client:
            return

        retcodes = []
        errors = []

        try:
            if self.options.subset:
                cmd_func = self.local_client.cmd_subset
                kwargs["subset"] = self.options.subset
                kwargs["cli"] = True
            else:
                cmd_func = self.local_client.cmd_cli

            if self.options.progress:
                kwargs["progress"] = True
                self.config["progress"] = True
                ret = {}
                for progress in cmd_func(**kwargs):
                    out = "progress"
                    try:
                        self._progress_ret(progress, out)
                    except LoaderError as exc:
                        raise SaltSystemExit(exc)
                    if "return_count" not in progress:
                        ret.update(progress)
                self._progress_end(out)
                self._print_returns_summary(ret)
            elif self.config["fun"] == "sys.doc":
                ret = {}
                out = ""
                for full_ret in self.local_client.cmd_cli(**kwargs):
                    ret_, out, retcode = self._format_ret(full_ret)
                    ret.update(ret_)
                self._output_ret(ret, out, retcode=retcode)
            else:
                if self.options.verbose:
                    kwargs["verbose"] = True
                ret = {}
                for full_ret in cmd_func(**kwargs):
                    try:
                        ret_, out, retcode = self._format_ret(full_ret)
                        retcodes.append(retcode)
                        self._output_ret(ret_, out, retcode=retcode)
                        ret.update(full_ret)
                    except KeyError:
                        errors.append(full_ret)

            # Returns summary
            if self.config["cli_summary"] is True:
                if self.config["fun"] != "sys.doc":
                    if self.options.output is None:
                        self._print_returns_summary(ret)
                        self._print_errors_summary(errors)

            # NOTE: Return code is set here based on if all minions
            # returned 'ok' with a retcode of 0.
            # This is the final point before the 'salt' cmd returns,
            # which is why we set the retcode here.
            if not all(
                exit_code == salt.defaults.exitcodes.EX_OK for exit_code in retcodes
            ):
                sys.stderr.write("ERROR: Minions returned with non-zero exit code\n")
                sys.exit(salt.defaults.exitcodes.EX_GENERIC)

        except (
            AuthenticationError,
            AuthorizationError,
            SaltInvocationError,
            EauthAuthenticationError,
            SaltClientError,
        ) as exc:
            ret = str(exc)
            self._output_ret(ret, "", retcode=1)
        finally:
            self.local_client.destroy()

    def _preview_target(self):
        """
        Return a list of minions from a given target
        """
        return self.local_client.gather_minions(
            self.config["tgt"], self.selected_target_option or "glob"
        )

    def _run_batch(self):
        import salt.cli.batch

        eauth = {}
        if "token" in self.config:
            eauth["token"] = self.config["token"]

        # If using eauth and a token hasn't already been loaded into
        # kwargs, prompt the user to enter auth credentials
        if "token" not in eauth and self.options.eauth:
            # This is expensive. Don't do it unless we need to.
            import salt.auth

            resolver = salt.auth.Resolver(self.config)
            res = resolver.cli(self.options.eauth)
            if self.options.mktoken and res:
                tok = resolver.token_cli(self.options.eauth, res)
                if tok:
                    eauth["token"] = tok.get("token", "")
            if not res:
                sys.stderr.write("ERROR: Authentication failed\n")
                sys.exit(2)
            eauth.update(res)
            eauth["eauth"] = self.options.eauth

        if self.options.static:

            if not self.options.batch:
                self.config["batch"] = "100%"

            try:
                batch = salt.cli.batch.Batch(self.config, eauth=eauth, quiet=True)
            except SaltClientError:
                sys.exit(2)

            ret = {}

            for res, _ in batch.run():
                ret.update(res)

            self._output_ret(ret, "")

        else:
            try:
                self.config["batch"] = self.options.batch
                batch = salt.cli.batch.Batch(
                    self.config, eauth=eauth, _parser=self.options
                )
            except SaltClientError:
                # We will print errors to the console further down the stack
                sys.exit(1)
            # Printing the output is already taken care of in run() itself
            retcode = 0
            for res, job_retcode in batch.run():
                if job_retcode > retcode:
                    # Exit with the highest retcode we find
                    retcode = job_retcode
            sys.exit(retcode)

    def _print_errors_summary(self, errors):
        if errors:
            salt.utils.stringutils.print_cli("\n")
            salt.utils.stringutils.print_cli("---------------------------")
            salt.utils.stringutils.print_cli("Errors")
            salt.utils.stringutils.print_cli("---------------------------")
            for error in errors:
                salt.utils.stringutils.print_cli(self._format_error(error))

    def _print_returns_summary(self, ret):
        """
        Display returns summary
        """
        return_counter = 0
        not_return_counter = 0
        not_return_minions = []
        not_response_minions = []
        not_connected_minions = []
        failed_minions = []
        for each_minion in ret:
            minion_ret = ret[each_minion]
            if isinstance(minion_ret, dict) and "ret" in minion_ret:
                minion_ret = ret[each_minion].get("ret")
            if isinstance(minion_ret, str) and minion_ret.startswith(
                "Minion did not return"
            ):
                if "Not connected" in minion_ret:
                    not_connected_minions.append(each_minion)
                elif "No response" in minion_ret:
                    not_response_minions.append(each_minion)
                not_return_counter += 1
                not_return_minions.append(each_minion)
            else:
                return_counter += 1
                if self._get_retcode(ret[each_minion]):
                    failed_minions.append(each_minion)
        salt.utils.stringutils.print_cli("\n")
        salt.utils.stringutils.print_cli("-------------------------------------------")
        salt.utils.stringutils.print_cli("Summary")
        salt.utils.stringutils.print_cli("-------------------------------------------")
        salt.utils.stringutils.print_cli(
            f"# of minions targeted: {return_counter + not_return_counter}"
        )
        salt.utils.stringutils.print_cli(f"# of minions returned: {return_counter}")
        salt.utils.stringutils.print_cli(
            f"# of minions that did not return: {not_return_counter}"
        )
        salt.utils.stringutils.print_cli(
            f"# of minions with errors: {len(failed_minions)}"
        )
        if self.options.verbose:
            if not_connected_minions:
                salt.utils.stringutils.print_cli(
                    "Minions not connected: {}".format(" ".join(not_connected_minions))
                )
            if not_response_minions:
                salt.utils.stringutils.print_cli(
                    "Minions not responding: {}".format(" ".join(not_response_minions))
                )
            if failed_minions:
                salt.utils.stringutils.print_cli(
                    "Minions with failures: {}".format(" ".join(failed_minions))
                )
        salt.utils.stringutils.print_cli("-------------------------------------------")

    def _progress_end(self, out):
        import salt.output

        salt.output.progress_end(self.progress_bar)

    def _progress_ret(self, progress, out):
        """
        Print progress events
        """
        import salt.output

        # Get the progress bar
        if not hasattr(self, "progress_bar"):
            try:
                self.progress_bar = salt.output.get_progress(self.config, out, progress)
            except Exception:  # pylint: disable=broad-except
                raise LoaderError(
                    "\nWARNING: Install the `progressbar` python package. "
                    "Requested job was still run but output cannot be displayed.\n"
                )
        salt.output.update_progress(self.config, progress, self.progress_bar, out)

    def _output_ret(self, ret, out, retcode=0):
        """
        Print the output from a single return to the terminal
        """
        import salt.output

        # Handle special case commands
        if self.config["fun"] == "sys.doc" and not isinstance(ret, Exception):
            self._print_docs(ret)
        else:
            # Determine the proper output method and run it
            salt.output.display_output(ret, out=out, opts=self.config, _retcode=retcode)
        if not ret:
            sys.stderr.write("ERROR: No return received\n")
            sys.exit(2)

    def _format_ret(self, full_ret):
        """
        Take the full return data and format it to simple output
        """
        ret = {}
        out = ""
        retcode = 0
        for key, data in full_ret.items():
            ret[key] = data["ret"]
            if "out" in data:
                out = data["out"]
            ret_retcode = self._get_retcode(data)
            if ret_retcode > retcode:
                retcode = ret_retcode
        return ret, out, retcode

    def _get_retcode(self, ret):
        """
        Determine a retcode for a given return
        """
        retcode = 0
        # if there is a dict with retcode, use that
        if isinstance(ret, dict) and ret.get("retcode", 0) != 0:
            if isinstance(ret.get("retcode", 0), dict):
                return max(ret.get("retcode", {0: 0}).values())
            return ret["retcode"]
        # if its a boolean, False means 1
        elif isinstance(ret, bool) and not ret:
            return 1
        return retcode

    def _format_error(self, minion_error):
        for minion, error_doc in minion_error.items():
            error = "Minion [{}] encountered exception '{}'".format(
                minion, error_doc["message"]
            )
        return error

    def _print_docs(self, ret):
        """
        Print out the docstrings for all of the functions on the minions
        """
        import salt.output

        docs = {}
        if not ret:
            self.exit(2, "No minions found to gather docs from\n")
        if isinstance(ret, str):
            self.exit(2, f"{ret}\n")
        for host in ret:
            if isinstance(ret[host], str) and (
                ret[host].startswith("Minion did not return")
                or ret[host] == "VALUE_TRIMMED"
            ):
                continue
            for fun in ret[host]:
                if fun not in docs and ret[host][fun]:
                    docs[fun] = ret[host][fun]
        if self.options.output:
            for fun in sorted(docs):
                salt.output.display_output({fun: docs[fun]}, "nested", self.config)
        else:
            for fun in sorted(docs):
                salt.utils.stringutils.print_cli(f"{fun}:")
                salt.utils.stringutils.print_cli(docs[fun])
                salt.utils.stringutils.print_cli("")



"""
Tests for win_file execution module
"""

import pytest

import salt.modules.win_file as win_file
import salt.utils.path

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture
def configure_loader_modules():
    return {win_file: {"__utils__": {"path.islink": salt.utils.path.islink}}}


@pytest.fixture(scope="function")
def test_dir(tmp_path_factory):
    test_dir = tmp_path_factory.mktemp("test_dir")
    yield test_dir
    if test_dir.exists():
        test_dir.rmdir()


def test_issue_52002_check_file_remove_symlink(test_dir):
    """
    Make sure that directories including symlinks or symlinks can be removed
    """
    # Create environment
    target = test_dir / "child 1" / "target"
    target.mkdir(parents=True)

    symlink = test_dir / "child 2" / "link"
    symlink.parent.mkdir(parents=True)
    symlink.symlink_to(target, target_is_directory=True)

    # Test removal of directory containing symlink
    win_file.remove(test_dir)
    assert test_dir.exists() is False



"""
    Test for salt.modules.vmctl
"""

import pytest

import salt.modules.vmctl as vmctl
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {vmctl: {}}


def test_create_disk():
    """
    Tests creating a new disk image.
    """
    ret = {}
    ret["stdout"] = "vmctl: imagefile created"
    ret["stderr"] = ""
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
        assert vmctl.create_disk("/path/to/disk.img", "1G")


def test_load():
    """
    Tests loading a configuration file.
    """
    ret = {}
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
        assert vmctl.load("/etc/vm.switches.conf")


def test_reload():
    """
    Tests reloading the configuration.
    """
    ret = {}
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
        assert vmctl.reload()


def test_reset():
    """
    Tests resetting VMM.
    """
    ret = {}
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
        res = vmctl.reset()
        mock_cmd.assert_called_once_with(
            ["vmctl", "reset"], output_loglevel="trace", python_shell=False
        )
        assert res


def test_reset_vms():
    """
    Tests resetting VMs.
    """
    ret = {}
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
        res = vmctl.reset(vms=True)
        mock_cmd.assert_called_once_with(
            ["vmctl", "reset", "vms"], output_loglevel="trace", python_shell=False
        )
        assert res


def test_reset_switches():
    """
    Tests resetting switches.
    """
    ret = {}
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
        res = vmctl.reset(switches=True)
        mock_cmd.assert_called_once_with(
            ["vmctl", "reset", "switches"],
            output_loglevel="trace",
            python_shell=False,
        )
        assert res


def test_reset_all():
    """
    Tests resetting all.
    """
    ret = {}
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
        res = vmctl.reset(all=True)
        mock_cmd.assert_called_once_with(
            ["vmctl", "reset", "all"], output_loglevel="trace", python_shell=False
        )
        assert res


def test_start_existing_vm():
    """
    Tests starting a VM that is already defined.
    """
    ret = {}
    ret["stderr"] = "vmctl: started vm 4 successfully, tty /dev/ttyp4"
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    expected = {"changes": True, "console": "/dev/ttyp4"}
    with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
        assert vmctl.start("4") == expected


def test_start_new_vm():
    """
    Tests starting a new VM.
    """
    ret = {}
    ret["stderr"] = "vmctl: started vm 4 successfully, tty /dev/ttyp4"
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    mock_status = MagicMock(return_value={})
    expected = {"changes": True, "console": "/dev/ttyp4"}
    with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
        with patch("salt.modules.vmctl.status", mock_status):
            res = vmctl.start("web1", bootpath="/bsd.rd", nics=2, disk="/disk.img")
            mock_cmd.assert_called_once_with(
                [
                    "vmctl",
                    "start",
                    "web1",
                    "-i 2",
                    "-b",
                    "/bsd.rd",
                    "-d",
                    "/disk.img",
                ],
                output_loglevel="trace",
                python_shell=False,
            )
            assert res == expected


def test_status():
    """
    Tests getting status for all VMs.
    """
    ret = {}
    ret["stdout"] = (
        "   ID   PID VCPUS  MAXMEM  CURMEM     TTY        OWNER NAME\n"
        "    1   123     1    2.9G     150M  ttyp5       john web1 - stopping\n"
        "    2   456     1    512M     301M  ttyp4       paul web2\n"
        "    3     -     1    512M       -       -       george web3\n"
    )
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    expected = {
        "web1": {
            "curmem": "150M",
            "id": "1",
            "maxmem": "2.9G",
            "owner": "john",
            "pid": "123",
            "state": "stopping",
            "tty": "ttyp5",
            "vcpus": "1",
        },
        "web2": {
            "curmem": "301M",
            "id": "2",
            "maxmem": "512M",
            "owner": "paul",
            "pid": "456",
            "state": "running",
            "tty": "ttyp4",
            "vcpus": "1",
        },
        "web3": {
            "curmem": "-",
            "id": "3",
            "maxmem": "512M",
            "owner": "george",
            "pid": "-",
            "state": "stopped",
            "tty": "-",
            "vcpus": "1",
        },
    }
    with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
        assert vmctl.status() == expected


def test_status_single():
    """
    Tests getting status for a single VM.
    """
    ret = {}
    ret["stdout"] = (
        "   ID   PID VCPUS  MAXMEM  CURMEM     TTY        OWNER NAME\n"
        "    1   123     1    2.9G     150M  ttyp5       ringo web4\n"
        "    2     -     1    512M       -       -       george web3\n"
    )
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    expected = {
        "web4": {
            "curmem": "150M",
            "id": "1",
            "maxmem": "2.9G",
            "owner": "ringo",
            "pid": "123",
            "state": "running",
            "tty": "ttyp5",
            "vcpus": "1",
        },
    }
    with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
        assert vmctl.status("web4") == expected


def test_stop_when_running():
    """
    Tests stopping a VM that is running.
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "vmctl: sent request to terminate vm 14"
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
        res = vmctl.stop("web1")
        mock_cmd.assert_called_once_with(
            ["vmctl", "stop", "web1"], output_loglevel="trace", python_shell=False
        )
        assert res["changes"]


def test_stop_when_stopped():
    """
    Tests stopping a VM that is already stopped/stopping.
    """
    ret = {}
    ret["stdout"] = ""
    ret["stderr"] = "vmctl: terminate vm command failed: Invalid argument"
    ret["retcode"] = 0
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(vmctl.__salt__, {"cmd.run_all": mock_cmd}):
        res = vmctl.stop("web1")
        mock_cmd.assert_called_once_with(
            ["vmctl", "stop", "web1"], output_loglevel="trace", python_shell=False
        )
        assert not res["changes"]
